""" Base module for authenticating to the aWhere API. """

import base64
import requests as rq


def get_oauth_token(key, secret):
    """Returns an OAuth Token used to authenticate to the
    aWhere API if a valid key and secret are provided.

    API reference: https://docs.awhere.com/knowledge-base-docs/authentication/

    Parameters
    ----------
    key : str
        API key for a valid aWhere API application.

    secret : str
        API secret for a valid aWhere API application.

    Returns
    -------
    access_token : str
        Access token created from the API key and secret.

    Example
    -------
        >>> # Imports
        >>> import os
        >>> import awherepy as aw
        >>> # Get aWhere API key and secret
        >>> awhere_api_key = os.environ.get('AWHERE_API_KEY')
        >>> awhere_api_secret = os.environ.get('AWHERE_API_SECRET')
        >>> # Get OAuth token
        >>> auth_token = aw.get_oauth_token(awhere_api_key, awhere_api_secret)
    """
    # Base64 Encode the API key/secret pair
    encoded_key_secret = base64.b64encode(
        bytes(f"{key}:{secret}", "utf-8")
    ).decode("ascii")

    # Define authorization parameters for HTTP request
    auth_url = "https://api.awhere.com/oauth/token"

    auth_headers = {
        "Authorization": f"Basic {encoded_key_secret}",
        "Content-Type": "application/x-www-form-urlencoded",
    }

    auth_body = "grant_type=client_credentials"

    # Request OAuth Token
    response = rq.post(url=auth_url, headers=auth_headers, data=auth_body)

    # Get access token
    access_token = response.json().get("access_token")

    # Return access token
    return access_token


def valid_credentials(key, secret):
    """Returns True if aWhere API credentials are valid, else False.

    Parameters
    ----------
    key : str
        API key for a valid aWhere API application.

    secret : str
        API secret for a valid aWhere API application.

    Returns
    -------
    valid : bool
        Boolean indicating True if API credentials are valid, else false.

    Example
    -------
        >>> # Imports
        >>> import os
        >>> import awherepy as aw
        >>> # Get aWhere API key and secret
        >>> awhere_api_key = os.environ.get('AWHERE_API_KEY')
        >>> awhere_api_secret = os.environ.get('AWHERE_API_SECRET')
        >>> # Check validity of credentials and get OAuth token
        >>> if aw.valid_credentials(awhere_api_key, awhere_api_secret):
        ...     auth_token = aw.get_oauth_token(
        ...         awhere_api_key, awhere_api_secret
        ...     )
    """
    # Get OAuth token
    token = get_oauth_token(key, secret)

    # Check for validity
    valid = False if token is None else True

    # Return valid
    return valid
    # return False if get_oauth_token(key, secret) is None else True



"""
awherepy.plantings
==================
A module to access and retrieve aWhere agronomic planting data.
"""

import requests as rq
from pandas.io.json import json_normalize
import awherepy as aw
import awherepy.fields as awf

# Define variables for data cleaning
PLANTINGS_DROP_COLS = [
    "_links.self.href",
    "_links.curies",
    "_links.awhere:crop.href",
    "_links.awhere:field.href",
]

PLANTINGS_RENAME_MAP = {
    "id": "planting_id",
    "crop": "crop_id",
    "field": "field_id",
    "plantingDate": "planting_date",
    "harvestDate": "harvest_date_actual",
    "yield.amount": "yield_amount_actual",
    "yield.units": "yield_amount_actual_units",
    "projections.yield.amount": "yield_amount_projected",
    "projections.yield.units": "yield_amount_projected_units",
    "projections.harvestDate": "harvest_date_projected",
}


def create_planting(key, secret, field_id, planting_info):
    """Creates a planting in an aWhere field.

    API reference: https://docs.awhere.com/knowledge-base-docs/field-plantings/

    Parameters
    ----------
    key : str
        API key for a valid aWhere API application.

    secret : str
        API secret for a valid aWhere API application.

    field_id : str
        Field ID for an existing aWhere field.

    planting_info : dict
        Dictionary containing the information required to create an aWhere
        planting. Can contain the following keys:

            crop: str, required
                ID for the crop that will be planted. Must be a valid aWhere
                crop ID.

            planting_date: str, required
                Planting date for the crop, formatted as 'YYYY-MM-DD'.

            projected_yield_amount: str or int or float, optional
                Projected yield amount for the crop. If used, the
                projected_yield_units parameter must also be set.

            projected_yield_units: str, optional
                Units for the projected crop yield. If used, the
                projected_yield_amount parameter must also be set.

            projected_harvest_date: str, optional
                Projected harvest date for the crop, formatted as
                'YYYY-MM-DD'.

            yield_amount: str or int or float, optional
                Actual yield amount for the crop. If used, the
                yield_units parameter must also be set.

            yield_units: str, optional
                Units for the actual crop yield. If used, the
                yield_amount parameter must also be set.

            harvest_date: str, optional
                Actual harvest date for the crop, formatted as 'YYYY-MM-DD'.

    Returns
    -------
    planting : pandas dataframe
        Dataframe containing the planting information.

    Example
    -------
        >>> # Imports
        >>> import os
        >>> import awherepy as aw
        >>> import ahwerepy.plantings as awp
        >>> # Get aWhere API key and secret
        >>> awhere_api_key = os.environ.get('AWHERE_API_KEY')
        >>> awhere_api_secret = os.environ.get('AWHERE_API_SECRET')
        >>> # Create planting for Manchester, Vermont field
        >>> vt_planting_info = {
        ...     'crop': 'potato-generic',
        ...     'planting_date': '2020-05-01',
        ...     'projected_harvest_date': '2020-09-30',
        ...     'projected_yield_amount': 200,
        ...     'projected_yield_units': 'boxes'
        ... }
        >>> awp.create_planting(
        ...     awhere_api_key,
        ...     awhere_api_secret,
        ...     field_id='VT-Manchester',
        ...     planting_info=vt_planting_info
        ... )
        Attempting to create planting...
        Created planting: potato-generic planted in VT-Manchester

                            crop_id         field_id    planting_date
        planting_id
        # potato-generic    VT-Manchester       2020-05-01
    """
    # Check if credentials are valid
    if aw.valid_credentials(key, secret):

        # Check planting_info object type
        if not isinstance(planting_info, dict):
            raise TypeError(
                "Invalid type: 'planting_info' must be of type dictionary."
            )

        # Raise errors for missing required parameters
        if not planting_info.get("crop"):
            raise KeyError("Missing required planting parameter: 'crop'.")

        if not planting_info.get("planting_date"):
            raise KeyError(
                "Missing required planting parameter: 'planting_date'."
            )

        if planting_info.get(
            "projected_yield_amount"
        ) and not planting_info.get("projected_yield_units"):
            raise KeyError(
                "Missing required planting parameter: 'projected_yield_units'."
            )

        if planting_info.get(
            "projected_yield_units"
        ) and not planting_info.get("projected_yield_amount"):
            raise KeyError(
                (
                    "Missing required planting parameter: "
                    "'projected_yield_amount'."
                )
            )

        if planting_info.get("yield_amount") and not planting_info.get(
            "yield_units"
        ):
            raise KeyError(
                "Missing required planting parameter: 'yield_units'."
            )

        if planting_info.get("yield_units") and not planting_info.get(
            "yield_amount"
        ):
            raise KeyError(
                "Missing required planting parameter: 'yield_amount'."
            )

        # Define plantings api url
        api_url = (
            f"https://api.awhere.com/v2/agronomics/fields/{field_id}/plantings"
        )

        # Get OAuth token
        auth_token = aw.get_oauth_token(key, secret)

        # Setup the HTTP request headers
        auth_headers = {
            "Authorization": f"Bearer {auth_token}",
            "Content-Type": "application/json",
        }

        # Define request body
        field_body = {
            "crop": planting_info.get("crop"),
            "plantingDate": planting_info.get("planting_date"),
            "projections": {
                "yield": {
                    "amount": planting_info.get("projected_yield_amount"),
                    "units": planting_info.get("projected_yield_units"),
                },
                "harvestDate": planting_info.get("projected_harvest_date"),
            },
            "yield": {
                "amount": planting_info.get("yield_amount"),
                "units": planting_info.get("yield_units"),
            },
            "harvestDate": planting_info.get("harvest_date"),
        }

        # Perform the POST request to create Field
        print("Attempting to create planting...")
        response = rq.post(api_url, headers=auth_headers, json=field_body)

        # Check if request succeeded
        if response.ok:

            # Get planting for output
            planting = get_plantings(
                key,
                secret,
                kwargs={
                    "field_id": planting_info.get("field_id"),
                    "planting_id": "current",
                },
            )

            # Indicate success
            print(
                (
                    f"Created planting: {planting_info.get('crop')} "
                    f"planted in {field_id}"
                )
            )

        else:

            # Indicate error
            print("Failed to create field.")

    # Invalid credentials
    else:
        # Raise error
        raise ValueError("Invalid aWhere API credentials.")

    # Return planting
    return planting


def get_plantings(key, secret, kwargs=None):
    """Gets all aWhere plantings associated an account, all plantings
    associated with a with a specified field, or an individual planting.

    API reference: https://docs.awhere.com/knowledge-base-docs/field-plantings/

    Parameters
    ----------
    key : str
        API key for a valid aWhere API application.

    secret : str
        API secret for a valid aWhere API application.

    kwargs : dict, optional
        Keyword arguments for different query parameters.
        Running the function without kwargs will use the
        default values. Arguments include:

            field_id: str
                Field ID for a valid aWhere field associated with the input API
                key/secret. If not specified and planting_id not specified,
                all plantings for all fields will be returned. If specified and
                planting_id not specified, all plantings associated with the
                specific field will be returned.

            planting_id: int
                Planting ID for a valid planting associated with an existing
                aWhere field. If specified and field_specified, a single
                planting will be returned. If not specified and field_id
                specified, the current (most recent) planting for that field
                will be returned.

            limit: int
                Number of results in the API response per page. Used with
                offset kwarg to sort through pages of results (cases where the
                number of results exceeds the limit per page). Applicable when
                the number of results exceeds 1. Maximum value is 10. Default
                value is 10.

            offset: int
                Number of results in the API response to skip. Used with limit
                kwarg to sorts through pages of results (cases where the
                number of results exceeds the limit per page). Applicable when
                the number of results exceeds 1. Default value is 0 (start
                with first result).

    Returns
    -------
    plantings_df : pandas dataframe
        Dataframe containing all plantings or an individual planting.

    Example
    -------
        >>> # Imports
        >>> import os
        >>> import awherepy as aw
        >>> import ahwerepy.plantings as awp
        >>> # Get aWhere API key and secret
        >>> awhere_api_key = os.environ.get('AWHERE_API_KEY')
        >>> awhere_api_secret = os.environ.get('AWHERE_API_SECRET')
        >>> # Get all plantings in all fields
        >>> all_plantings = awp.get_plantings(
        ...     awhere_api_key, awhere_api_secret
        ... )
        >>> # Get all plantings for a specific field
        >>> vt_plantings = awp.get_plantings(
        ...     awhere_api_key,
        ...     awhere_api_secret,
        ...     kwargs={'field_id': 'VT-Manchester'}
        ... )
        >>> # Get most recent planting (looks through all fields)
        >>> current_planting = awp.get_plantings(
        ...     awhere_api_key,
        ...     awhere_api_secret,
        ...     kwargs={'planting_id': 'current'}
        ... )
        >>> # Get specific planting
        >>> vt_planting = awp.get_plantings(
        ...     awhere_api_key,
        ...     awhere_api_secret,
        ...     kwargs={'planting_id': ######}
        ... )
    """
    # Define global variables
    global PLANTINGS_DROP_COLS, PLANTINGS_RENAME_MAP

    # Check if credentials are valid
    if aw.valid_credentials(key, secret):

        # Get OAuth token
        auth_token = aw.get_oauth_token(key, secret)

        # Setup the HTTP request headers
        auth_headers = {"Authorization": f"Bearer {auth_token}"}

        # All plantings from all fields
        if not kwargs:

            # Set limit and offset query parameters
            kwargs = {"limit": 10, "offset": 0}

            # Define api url
            api_url = (
                "https://api.awhere.com/v2/agronomics/plantings?limit="
                f"{kwargs.get('limit')}&offset={kwargs.get('offset')}"
            )

        # All plantings from single field
        elif kwargs.get("field_id") and not kwargs.get("planting_id"):

            # Set limit and offset query parameters if not existing
            if "limit" not in kwargs.keys():
                kwargs["limit"] = 10

            if "offset" not in kwargs.keys():
                kwargs["offset"] = 0

            # Define api url
            api_url = (
                "https://api.awhere.com/v2/agronomics/fields/"
                f"{kwargs.get('field_id')}/plantings"
            )

        # Single planting (id/current), field-independent
        elif kwargs.get("planting_id") and not kwargs.get("field_id"):

            # Define api url
            api_url = (
                "https://api.awhere.com/v2/agronomics/plantings/"
                f"{kwargs.get('planting_id')}"
            )

        # Invalid combination
        else:

            # Raise error
            raise KeyError("Invalid kwarg combination.")

        # Get API response
        response = rq.get(f"{api_url}", headers=auth_headers)

        # Convert API response to JSON format
        response_json = response.json()

        # Convert to dataframe
        plantings_df = (
            json_normalize(response_json.get("plantings"))
            if response_json.get("plantings")
            else json_normalize(response_json)
        )

        # Drop unnecessary columns
        plantings_df.drop(
            columns=PLANTINGS_DROP_COLS, inplace=True, errors="ignore"
        )

        # Rename columns
        plantings_df.rename(
            columns=PLANTINGS_RENAME_MAP, inplace=True, errors="ignore"
        )

        # Set index
        plantings_df.set_index("planting_id", inplace=True)

    # Invalid credentials
    else:
        # Raise error
        raise ValueError("Invalid aWhere API credentials.")

    # Return the plantings dataframe
    return plantings_df


def update_planting(
    key, secret, planting_info,
):
    """Updates a specified planting.

    API reference: https://docs.awhere.com/knowledge-base-docs/field-plantings/

    Parameters
    ----------
    key : str
        API key for a valid aWhere API application.

    secret : str
        API secret for a valid aWhere API application.

    planting_info : dict
        Dictionary containing the information to update an aWhere planting.
        Can contain the following keys:

            field_id: str
                Field ID for an existing aWhere field.

            planting_id: int
                Planting ID for an existing planting.

            update_type : str
                The type of update. Options are 'full' or 'partial'. Full
                update replaces every parameter in the planting. Partial
                update replaces only the provided parameters in the planting.

            crop: str
                ID for the crop that will be planted. Must be a valid aWhere
                crop ID.

            planting_date: str, required
                Planting date for the crop, formatted as 'YYYY-MM-DD'.

            projected_yield_amount: str or int or float, optional
                Projected yield amount for the crop. If used, the
                projected_yield_units parameter must also be set.

            projected_yield_units: str, optional
                Units for the projected crop yield. If used, the
                projected_yield_amount parameter must also be set.

            projected_harvest_date: str, optional
                Projected harvest date for the crop, formatted as
                'YYYY-MM-DD'.

            yield_amount: str or int or float, optional
                Actual yield amount for the crop. If used, the
                yield_units parameter must also be set.

            yield_units: str, optional
                Units for the actual crop yield. If used, the
                yield_amount parameter must also be set.

            harvest_date: str, optional
                Actual harvest date for the crop, formatted as 'YYYY-MM-DD'.

    Example
    -------
        >>> # Imports
        >>> import os
        >>> import awherepy as aw
        >>> import ahwerepy.plantings as awp
        >>> # Get aWhere API key and secret
        >>> awhere_api_key = os.environ.get('AWHERE_API_KEY')
        >>> awhere_api_secret = os.environ.get('AWHERE_API_SECRET')
        >>> # Define planting update information
        >>> planting_update_info = {
        ...     'field_id': 'VT-Manchester',
        ...     'planting_id': 475994,
        ...     'update_type': 'partial'
        ...     'crop': 'sugarbeet-generic',
        ...     'planting_date': '2020-06-05',
        ...     'projections_yield_amount': 50,
        ...     'projections_yield_units': 'large boxes',
        ...     'projected_harvest_date': '2020-08-08',
        ...     'yield_amount': 200,
        ...     'yield_units': 'large boxes',
        ...     'harvest_date': '2020-07-31'
        ... }
        >>> # Update planting
        >>> awp.update_planting(
        ...     awhere_api_key,
        ...     awhere_api_secret,
        ...     planting_info=planting_update_info
        ... )
        Attempting to update planting...
        Updated planting: sugarbeet-generic in VT-Manchester
    """
    # Check if credentials are valid
    if aw.valid_credentials(key, secret):

        # Check planting_info object type
        if not isinstance(planting_info, dict):
            raise TypeError(
                "Invalid type: 'planting_info' must be of type dictionary."
            )

        # Raise error if field does not exist
        if (
            planting_info.get("field_id")
            not in awf.get_fields(key, secret).index
        ):
            raise KeyError("Field does not exist within account.")

        # Raise error if planting does not exist
        if planting_info.get("planting_id") != "current":
            if (
                planting_info.get("planting_id")
                not in get_plantings(key, secret).index
            ):
                raise KeyError("Planting does not exist within account.")

        # Define api url
        api_url = (
            "https://api.awhere.com/v2/agronomics/fields/"
            f"{planting_info.get('field_id')}/plantings/"
            f"{planting_info.get('planting_id')}"
        )

        # Get OAuth token
        auth_token = aw.get_oauth_token(key, secret)

        # Set up the HTTP request headers
        auth_headers = {
            "Authorization": f"Bearer {auth_token}",
            "Content-Type": "application/json",
        }

        # Full update
        if planting_info.get("update_type").lower() == "full":

            # Define request body
            field_body = {
                "crop": planting_info.get("crop"),
                "plantingDate": planting_info.get("planting_date"),
                "projections": {
                    "yield": {
                        "amount": planting_info.get("projected_yield_amount"),
                        "units": planting_info.get("projected_yield_units"),
                    },
                    "harvestDate": planting_info.get("projected_harvest_date"),
                },
                "yield": {
                    "amount": planting_info.get("yield_amount"),
                    "units": planting_info.get("yield_units"),
                },
                "harvestDate": planting_info.get("harvest_date"),
            }

            # Perform HTTP request to update planting information
            print("Attempting to update planting...")
            response = rq.put(
                f"{api_url}", headers=auth_headers, json=field_body,
            )

        # Partial update
        elif planting_info.get("update_type").lower() == "partial":

            # Define field body
            field_body = [
                {"op": "replace", "path": f"/{key}", "value": f"{value}"}
                for key, value in planting_info.items()
            ]

            # Perform HTTP request to update planting information
            print("Attempting to update planting...")
            response = rq.patch(
                f"{api_url}", headers=auth_headers, json=field_body,
            )

        # Invalid update type
        else:
            raise ValueError(
                "Invalid update type. Must be 'full' or 'partial'."
            )

        # Check if request succeeded
        if response.ok:

            # Get planting for output
            planting = get_plantings(
                key,
                secret,
                kwargs={
                    "field_id": planting_info.get("field_id"),
                    "planting_id": planting_info.get("planting_id"),
                },
            )

            # Indicate success
            print(
                (
                    f"Updated planting: {planting_info.get('planting_id')}"
                    f"in {planting_info.get('field_id')}"
                )
            )

        else:
            # Indicate error
            planting = "Failed to update planting."
            print(planting)

    # Invalid credentials
    else:
        # Raise error
        raise ValueError("Invalid aWhere API credentials.")

    # Return planting
    return planting


def delete_planting(key, secret, field_id, planting_id="current"):
    """Deletes a specified aWhere planting.

    API reference: https://docs.awhere.com/knowledge-base-docs/field-plantings/

    Parameters
    ----------
    key : str
        API key for a valid aWhere API application.

    secret : str
        API secret for a valid aWhere API application.

    field_id : str
        Field ID for an existing aWhere field.

    planting_id : int, optional
        Planting ID for an existing planting. This identifies the planting that
        will be deleted from the specified aWhere field. Default value is
        'current', which references the most recent planting associated with
        the specified field.

    Returns
    -------
    message : str
        Output message indicating success or failure for the planting deletion.

    Example
    -------
    >>> # Imports
    >>> import os
    >>> import awherepy as aw
    >>> import awherepy.plantings as awp
    >>> # Get aWhere API key and secret
    >>> awhere_api_key = os.environ.get('AWHERE_API_KEY')
    >>> awhere_api_secret = os.environ.get('AWHERE_API_SECRET')
    >>> # Delete planting in Manchester, Vermont field
    >>> awp.delete_planting(
    ...     awhere_api_key,
    ...     awhere_api_secret,
    ...     field_id='VT-Manchester',
    ...     planting_id=######
    ... )
    Attempting to delete planting...
    Deleted planting: ###### in VT-Manchester
    """
    # Check if credentials are valid
    if aw.valid_credentials(key, secret):

        # Raise error if field does not exist
        if field_id not in awf.get_fields(key, secret).index:
            raise KeyError("Field does not exist within account.")

        # Raise error if planting does not exist
        if planting_id != "current":
            if planting_id not in get_plantings(key, secret).index:
                raise KeyError("Planting does not exist within account.")

        # Define api url
        api_url = (
            f"https://api.awhere.com/v2/agronomics/fields/"
            f"{field_id}/plantings/{planting_id}"
        )

        # Get OAuth token
        auth_token = aw.get_oauth_token(key, secret)

        # Setup the HTTP request headers
        auth_headers = {
            "Authorization": f"Bearer {auth_token}",
        }

        # Perform the POST request to delete the planting
        print("Attempting to delete planting...")
        rq.delete(f"{api_url}", headers=auth_headers)

        # Check if planting exists within account
        try:
            get_plantings(
                key,
                secret,
                kwargs={"field_id": field_id, "planting_id": planting_id},
            )

        # Catch error if planting does not exist (was deleted)
        except KeyError:
            message = f"Deleted planting: {planting_id} in {field_id}"
            print(message)

        # If delete did not work
        else:
            message = "Could not delete planting."
            print(message)

    # Invalid credentials
    else:
        # Raise error
        raise ValueError("Invalid aWhere API credentials.")

    return message



"""
awherepy.weather
================
A module to access and retrieve aWhere weather data.
"""

import requests as rq
import pandas as pd
from pandas.io.json import json_normalize
import geopandas as gpd
import awherepy as aw

# Define variables for data cleaning
# Weather norms
NORMS_COORD_COLS = ["location.longitude", "location.latitude"]

NORMS_DROP_COLS = [
    "meanTemp.units",
    "maxTemp.units",
    "minTemp.units",
    "precipitation.units",
    "solar.units",
    "dailyMaxWind.units",
    "averageWind.units",
    "_links.curies",
    "_links.awhere:field.href",
]

NORMS_RENAME_MAP = {
    "meanTemp.average": "mean_temp_avg_cels",
    "meanTemp.stdDev": "mean_temp_std_dev_cels",
    "maxTemp.average": "max_temp_avg_cels",
    "maxTemp.stdDev": "max_temp_std_dev_cels",
    "minTemp.average": "min_temp_avg_cels",
    "minTemp.stdDev": "min_temp_std_dev_cels",
    "precipitation.average": "precip_avg_mm",
    "precipitation.stdDev": "precip_std_dev_mm",
    "solar.average": "solar_avg_w_h_per_m2",
    "solar.stdDev": "solar_avg_std_dev_w_h_per_m2",
    "minHumidity.average": "min_humiduty_avg_%",
    "minHumidity.stdDev": "min_humidity_std_dev_%",
    "maxHumidity.average": "max_humiduty_avg_%",
    "maxHumidity.stdDev": "max_humidity_std_dev_%",
    "dailyMaxWind.average": "daily_max_wind_avg_m_per_sec",
    "dailyMaxWind.stdDev": "daily_max_wind_std_dev_m_per_sec",
    "averageWind.average": "average_wind_m_per_sec",
    "averageWind.stdDev": "average_wind_std_dev_m_per_sec",
    "location.fieldId": "field_id",
}

# Observed weather
OBSERVED_COORD_COLS = ["location.longitude", "location.latitude"]

OBSERVED_DROP_COLS = [
    "temperatures.units",
    "precipitation.units",
    "solar.units",
    "wind.units",
    "_links.curies",
    "_links.awhere:field.href",
]

OBSERVED_RENAME_MAP = {
    "temperatures.max": "temp_max_cels",
    "temperatures.min": "temp_min_cels",
    "precipitation.amount": "precip_amount_mm",
    "solar.amount": "solar_energy_w_h_per_m2",
    "relativeHumidity.average": "rel_humidity_avg_%",
    "relativeHumidity.max": "rel_humidity_max_%",
    "relativeHumidity.min": "rel_humidity_min_%",
    "wind.morningMax": "wind_morning_max_m_per_sec",
    "wind.dayMax": "wind_day_max_m_per_sec",
    "wind.average": "wind_avg_m_per_sec",
    "location.fieldId": "field_id",
}

# Weather forecast
FORECAST_COORD_COLS = ["longitude", "latitude"]

FORECAST_DROP_COLS = [
    "temperatures.units",
    "precipitation.units",
    "solar.units",
    "wind.units",
    "dewPoint.units",
    "units",
]

FORECAST_RENAME_MAP = {
    "startTime": "start_time",
    "endTime": "end_time",
    "conditionsCode": "conditions_code",
    "conditionsText": "conditions_text",
    "temperatures.value": "temp_val_cels",
    "temperatures.max": "temp_max_cels",
    "temperatures.min": "temp_min_cels",
    "precipitation.chance": "precip_chance_%",
    "precipitation.amount": "precip_amount_mm",
    "sky.cloudCover": "sky_cloud_cover_%",
    "sky.sunshine": "sky_sunshine_%",
    "solar.amount": "solar_energy_w_h_per_m2",
    "relativeHumidity.average": "rel_humidity_avg_%",
    "relativeHumidity.max": "rel_humidity_max_%",
    "relativeHumidity.min": "rel_humidity_min_%",
    "wind.average": "wind_avg_m_per_sec",
    "wind.max": "wind_max_m_per_sec",
    "wind.min": "wind_min_m_per_sec",
    "wind.bearing": "wind_bearing_deg",
    "wind.direction": "wind_direction_compass",
    "dewPoint.amount": "dew_point_cels",
    "average_temp": "soil_temp_avg_cels",
    "max_temp": "soil_temp_max_cels",
    "min_temp": "soil_temp_min_cels",
    "average_moisture": "soil_moisture_avg_%",
    "max_moisture": "soil_moisture_max_%",
    "min_moisture": "soil_moisture_min_%",
}


def _call_weather_norms(
    key,
    secret,
    input_type="location",
    location=(-105.648222, 40.313250),
    field_id=None,
    start_date="01-01",
    end_date=None,
    limit=10,
    offset=0,
):
    """Helper function that calls the aWhere API for weather norms.
    """
    # Check input data type
    # Location-based
    if input_type == "location":

        # Raise error if location is not defined
        if location is None:
            raise ValueError(
                ("Must specify a location, with longitude " "and latitude.")
            )

        # Set URL to location
        api_url = (
            "https://api.awhere.com/v2/weather/locations/"
            f"{location[1]},{location[0]}"
        )

    # Field-based
    elif input_type == "field":

        # Raise error if field name is not defined
        if field_id is None:
            raise ValueError("Must specify a field name.")

        # Set URL to fields
        api_url = f"https://api.awhere.com/v2/weather/fields/{field_id}"

    # Invalid input
    else:
        raise ValueError("Invalid input type. Must be 'location' or 'field'.")

    # Get OAuth token
    auth_token = aw.get_oauth_token(key, secret)

    # Set up the HTTP request headers
    auth_headers = {"Authorization": f"Bearer {auth_token}"}

    # Define URL variants - location
    url_single_day = (
        f"{api_url}/norms/{start_date}?limit={limit}&offset={offset}"
    )
    url_multiple_days = (
        f"{api_url}/norms/{start_date},{end_date}"
        f"?limit={limit}&offset={offset}"
    )

    # Get single-day or multi-day norms
    response = (
        rq.get(url_multiple_days, headers=auth_headers)
        if end_date
        else rq.get(url_single_day, headers=auth_headers)
    )

    # Convert response to json format
    weather_norms = response.json()

    # Return norms
    return weather_norms


def _extract_weather_norms(historic_norms):
    """Helper function that extracts weather norms into a dataframe.
    """
    # Raise error if input is not of type dictionary
    if not isinstance(historic_norms, dict):
        raise TypeError("Input data must be a dictionary.")

    # Check if multiple entries (days) are in norms
    if historic_norms.get("norms"):
        # Flatten to dataframe
        historic_norms_df = json_normalize(historic_norms.get("norms"))

    # Single-day norm
    else:
        # Flatten to dataframe
        historic_norms_df = json_normalize(historic_norms)

    # Set day as index
    historic_norms_df.set_index("day", inplace=True)

    # Drop unnecessary columns
    historic_norms_df.drop(columns=["_links.self.href"], axis=1, inplace=True)

    # Return dataframe
    return historic_norms_df


def _clean_weather_norms(df):
    """Helper function that cleans and georeferences weather norms.
    """
    # Define global variables
    global NORMS_COORD_COLS, NORMS_DROP_COLS, NORMS_RENAME_MAP

    # Define CRS (EPSG 4326)
    crs = "epsg:4326"

    # Create copy of input dataframe; prevents altering the original
    df_copy = df.copy()

    # Convert to geodataframe
    gdf = gpd.GeoDataFrame(
        df_copy,
        crs=crs,
        geometry=gpd.points_from_xy(
            df[NORMS_COORD_COLS[0]], df[NORMS_COORD_COLS[1]]
        ),
    )

    # Add lat/lon columns to drop columns list
    NORMS_DROP_COLS += NORMS_COORD_COLS

    # Drop columns
    gdf.drop(columns=NORMS_DROP_COLS, axis=1, inplace=True, errors="ignore")

    # Rename columns
    gdf.rename(columns=NORMS_RENAME_MAP, inplace=True, errors="ignore")

    # Return cleaned up geodataframe
    return gdf


def get_weather_norms(key, secret, kwargs=None):
    """Gets historical weather norms data from the aWhere API.

    API reference: https://docs.awhere.com/knowledge-base-docs/weather/

    Parameters
    ----------
    key : str
        API key for a valid aWhere API application.

    secret : str
        API secret for a valid aWhere API application.

    kwargs : dict, optional
        Keyword arguments for different query parameters.
        Running the function without kwargs will use the
        default values. Arguments include:

            input_type: str
                Type of data, weather API by geolocation or API by field.
                Valid options include 'location' and 'field'. Default
                value is 'location'.

            location: tuple of float
                Tuple containing the location (longitude, latitude) that the
                API gets data from. For use with the 'location' input_type.
                Default value is (-105.648222, 40.313250), for Bear Lake, CO.

            field_id: str
                Field ID for a valid aWhere field associated with the input API
                key/secret. For use with the 'field' input_type. Default value
                is None.

            start_date: str
                Start date for the weather norms data to be retrieved.
                Formatted 'MM-DD'. All days of the year are valid, from
                '01-01' to '12-31'. Default value is '01-01'.

            end_date: str
                End date for the weather norms data to be retrieved. Formatted
                'MM-DD'. All days of the year are valid, from '01-01' to
                '12-31'. Default value is None.

            limit: int
                Number of results in the API response per page. Used with
                offset kwarg to sort through pages of results (cases where the
                number of results exceeds the limit per page). Applicable when
                the number of results exceeds 1. Maximum value is 10. Default
                value is 10.

            offset: int
                Number of results in the API response to skip. Used with limit
                kwarg to sort through pages of results (cases where the
                number of results exceeds the limit per page). Applicable when
                the number of results exceeds 1. Default value is 0 (start
                with first result).

    Returns
    -------
    norms_gdf : geopandas geodataframe
        Geotdataframe containing the historical weather norms data for the
        specified location or aWhere field, for the specified date(s).

    Example
    -------
        >>> # Imports
        >>> import os
        >>> import awherepy as aw
        >>> import awherepy.weather as aww
        >>> # Get aWhere API key and secret
        >>> awhere_api_key = os.environ.get('AWHERE_API_KEY')
        >>> awhere_api_secret = os.environ.get('AWHERE_API_SECRET')
        >>> # Get historical norms with default kwargs (Bear Lake, CO)
        >>> bear_lake_norms = aww.get_weather_norms(
        ...     key=awhere_api_key, secret=awhere_api_secret)
        >>> # Check number of entries in geodataframe
        >>> len(bear_lake_norms)
        1
        >>> # Define non-default kwargs (Manchester Center, Vermont)
        >>> # Define kwargs for Manchester, Vermont
        >>> vt_kwargs = {
        ...     'location': (-73.0723269, 43.1636875),
        ...     'start_date': '05-10',
        ...     'end_date': '05-19'
        ... }
        >>> # Get historical norms for Manchester, Vermont
        >>> manchester_vt_norms = aww.get_weather_norms(
        ...     key=awhere_api_key,
        ...     secret=awhere_api_secret,
        ...     kwargs=vt_kwargs
        ... )
        >>> # Check number entries in geodataframe
        >>> len(manchester_vt_norms)
        10
    """
    # Check if credentials are valid
    if aw.valid_credentials(key, secret):

        # Call data
        norms_json = (
            _call_weather_norms(key, secret, **kwargs)
            if kwargs
            else _call_weather_norms(key, secret)
        )

        # Extract data
        norms_df = _extract_weather_norms(norms_json)

        # Clean data
        norms_gdf = _clean_weather_norms(norms_df)

    else:
        # Raise error
        raise ValueError("Invalid aWhere API credentials.")

    # Return cleaned data
    return norms_gdf


def _call_weather_observed(
    key,
    secret,
    input_type="location",
    location=(-105.648222, 40.313250),
    field_id=None,
    start_date=None,
    end_date=None,
    limit=10,
    offset=0,
):
    """Helper function that calls the aWhere API for observed weather.
    """
    # Check input data type
    # Location-based
    if input_type == "location":

        # Raise error if location is not defined
        if location is None:
            raise ValueError(
                ("Must specify a location, with longitude " "and latitude.")
            )

        # Set URL to location
        api_url = (
            "https://api.awhere.com/v2/weather/locations/"
            f"{location[1]},{location[0]}"
        )

    # Field-based
    elif input_type == "field":

        # Raise error if field name is not defined
        if field_id is None:
            raise ValueError("Must specify a field name.")

        # Set URL to fields
        api_url = f"https://api.awhere.com/v2/weather/fields/{field_id}"

    # Invalid input
    else:
        raise ValueError("Invalid input type. Must be 'location' or 'field'.")

    # Get OAuth token
    auth_token = aw.get_oauth_token(key, secret)

    # Set up the HTTP request headers
    auth_headers = {"Authorization": f"Bearer {auth_token}"}

    # Define URL variants
    url_no_date = f"{api_url}/observations?limit={limit}&offset={offset}"
    url_start_date = f"{api_url}/observations/{start_date}"
    url_end_date = f"{api_url}/observations/{end_date}"
    url_both_dates = (
        f"{api_url}/observations/{start_date},{end_date}"
        f"?limit={limit}&offset={offset}"
    )

    # Perform the HTTP request to obtain the norms for the Field
    # Default - 7-day
    if not (start_date or end_date):
        response = rq.get(url_no_date, headers=auth_headers)

    # Single date - specify start date
    elif start_date and not end_date:
        response = rq.get(url_start_date, headers=auth_headers)

    # Single date - specify end date
    elif end_date and not start_date:
        response = rq.get(url_end_date, headers=auth_headers)

    # Date range
    elif start_date and end_date:
        response = rq.get(url_both_dates, headers=auth_headers)

    # Convert response to json format
    weather_observed = response.json()

    # Return the observed
    return weather_observed


def _extract_weather_observed(observed_weather):
    """Helper function that extracts observed weather into a dataframe.
    """
    # Raise error if input is not of type dictionary
    if not isinstance(observed_weather, dict):
        raise TypeError("Input data must be a dictionary.")

    # Check if multiple entries (days) are in observed
    if observed_weather.get("observations"):
        # Flatten to dataframe
        observed_weather_df = json_normalize(
            observed_weather.get("observations")
        )

    # Single-day observed
    else:
        # Flatten to dataframe
        observed_weather_df = json_normalize(observed_weather)

    # Set date as index
    observed_weather_df.set_index("date", inplace=True)

    # Drop unnecessary columns
    observed_weather_df.drop(
        columns=["_links.self.href"], axis=1, inplace=True
    )

    # Return dataframe
    return observed_weather_df


def _clean_weather_observed(df):
    """Helper function that cleans and georeferences observed weather.
    """
    # Define global variables
    global OBSERVED_COORD_COLS, OBSERVED_DROP_COLS, OBSERVED_RENAME_MAP

    # Define CRS (EPSG 4326)
    crs = "epsg:4326"

    # Create copy of input dataframe; prevents altering the original
    df_copy = df.copy()

    # Convert to geodataframe
    gdf = gpd.GeoDataFrame(
        df_copy,
        crs=crs,
        geometry=gpd.points_from_xy(
            df[OBSERVED_COORD_COLS[0]], df[OBSERVED_COORD_COLS[1]]
        ),
    )

    # Add lat/lon columns to drop columns list
    OBSERVED_DROP_COLS += OBSERVED_COORD_COLS

    # Drop columns
    gdf.drop(columns=OBSERVED_DROP_COLS, axis=1, inplace=True, errors="ignore")

    # Rename columns
    gdf.rename(columns=OBSERVED_RENAME_MAP, inplace=True, errors="ignore")

    # Return cleaned up geodataframe
    return gdf


def get_weather_observed(key, secret, kwargs=None):
    """Gets observed weather data from the aWhere API.

    API reference: https://docs.awhere.com/knowledge-base-docs/weather/

    Parameters
    ----------
    key : str
        API key for a valid aWhere API application.

    secret : str
        API secret for a valid aWhere API application.

    kwargs : dict, optional
        Keyword arguments for different query parameters.
        Running the function without kwargs will use the
        default values. Arguments include:

            input_type: str
                Type of data, weather API by geolocation or API by field.
                Valid options include 'location' and 'field'. Default
                value is 'location'.

            location: tuple of float
                Tuple containing the location (longitude, latitude) that the
                API gets data from. For use with the 'location' input_type.
                Default value is (-105.648222, 40.313250), for Bear Lake, CO.

            field_id: str
                Field ID for a valid aWhere field associated with the input API
                key/secret. For use with the 'field' input_type. Default value
                is None.

            start_date: str
                Start date for the observed weather data to be retrieved.
                Formatted 'YYYY-MM-DD'. All days of the year are valid, from
                'YYYY-01-01' to 'YYYY-12-31'. Default value is None, which
                returns the previous 7 days of observed weather.

            end_date: str
                End date for the observed weather data to be retrieved.
                Formatted 'YYYY-MM-DD'. All days of the year are valid, from
                'YYYY-01-01' to 'YYYY-12-31'. Default value is None, which
                returns the previous 7 days of observed weather.

            limit: int
                Number of results in the API response per page. Used with
                offset kwarg to sort through pages of results (cases where the
                number of results exceeds the limit per page). Applicable when
                the number of results exceeds 1. Maximum value is 10. Default
                value is 10.

            offset: int
                Number of results in the API response to skip. Used with limit
                kwarg to sort through pages of results (cases where the
                number of results exceeds the limit per page). Applicable when
                the number of results exceeds 1. Default value is 0 (start
                with first result).

    Returns
    -------
    observed_gdf : geopandas geodataframe
        Geotdataframe containing the observed weather data for the
        specified location or aWhere field, for the specified date(s).

    Example
    -------
        >>> # Imports
        >>> import os
        >>> import awherepy as aw
        >>> import awherepy.weather as aww
        >>> # Get aWhere API key and secret
        >>> awhere_api_key = os.environ.get('AWHERE_API_KEY')
        >>> awhere_api_secret = os.environ.get('AWHERE_API_SECRET')
        >>> # Get observed weather with default kwargs (Bear Lake, CO)
        >>> bear_lake_observed = aww.get_weather_observed(
        ...     key=awhere_api_key, secret=awhere_api_secret)
        >>> # Check number of entries in geodataframe
        >>> len(bear_lake_observed)
        7
        >>> # Define non-default kwargs (Manchester Center, Vermont)
        >>> # Define kwargs for Manchester, Vermont
        >>> vt_kwargs = {
        ...     'location': (-73.0723269, 43.1636875),
        ...     'start_date': '2020-05-10',
        ...     'end_date': '2020-05-20'
        ... }
        >>> # Get observed weather for Manchester, Vermont
        >>> manchester_vt_observed = aww.get_weather_observed(
        ...     key=awhere_api_key,
        ...     secret=awhere_api_secret,
        ...     kwargs=vt_kwargs
        ... )
        >>> # Check number entries in geodataframe
        >>> len(manchester_vt_observed)
        10
    """
    # Check if credentials are valid
    if aw.valid_credentials(key, secret):

        # Call data
        observed_json = (
            _call_weather_observed(key, secret, **kwargs)
            if kwargs
            else _call_weather_observed(key, secret)
        )

        # Extract data
        observed_df = _extract_weather_observed(observed_json)

        # Clean data
        observed_gdf = _clean_weather_observed(observed_df)

    else:
        # Raise error
        raise ValueError("Invalid aWhere API credentials.")

    return observed_gdf


def _call_weather_forecast(
    key,
    secret,
    input_type="location",
    location=(-105.648222, 40.313250),
    field_id=None,
    start_date=None,
    end_date=None,
    limit=10,
    offset=0,
    block_size=24,
):
    """Helper function that calls the aWhere API for weather forecast data.
    """
    # Check input data type
    # Location-based
    if input_type == "location":

        # Raise error if location is not defined
        if location is None:
            raise ValueError(
                ("Must specify a location, with longitude " "and latitude.")
            )

        # Set URL to location
        api_url = (
            f"https://api.awhere.com/v2/weather/locations/"
            f"{location[1]},{location[0]}"
        )

    # Field-based
    elif input_type == "field":

        # Raise error if field name is not defined
        if field_id is None:
            raise ValueError("Must specify a field name.")

        # Set URL to fields
        api_url = f"https://api.awhere.com/v2/weather/fields/{field_id}"

    # Invalid input
    else:
        raise ValueError("Invalid input type. Must be 'location' or 'field'.")

    # Get OAuth token
    auth_token = aw.get_oauth_token(key, secret)

    # Setup the HTTP request headers
    auth_headers = {"Authorization": f"Bearer {auth_token}"}

    # Define URL variants
    url_no_date = (
        f"{api_url}/forecasts?limit={limit}&offset={offset}"
        f"&blockSize={block_size}"
    )
    url_start_date = (
        f"{api_url}/forecasts/{start_date}?limit={limit}"
        f"&offset={offset}&blockSize={block_size}"
    )
    url_end_date = (
        f"{api_url}/forecasts/{end_date}?limit={limit}"
        f"&offset={offset}&blockSize={block_size}"
    )
    url_both_dates = (
        f"{api_url}/forecasts/{start_date},{end_date}"
        f"?limit={limit}&offset={offset}&blockSize={block_size}"
    )

    # Perform the HTTP request to obtain the Forecast for the Field
    # Default - 7-day
    if not (start_date or end_date):
        response = rq.get(url_no_date, headers=auth_headers)

    # Single date - specify start day
    elif start_date and not end_date:
        response = rq.get(url_start_date, headers=auth_headers)

    # Single date - specify end day
    elif end_date and not start_date:
        response = rq.get(url_end_date, headers=auth_headers)

    # Date range
    elif start_date and end_date:
        response = rq.get(url_both_dates, headers=auth_headers)

    # Convert response to json format
    weather_forecast = response.json()

    # Return forecast
    return weather_forecast


def _extract_weather_forecast(forecast, forecast_type="main"):
    """Helper function that extracts weather foreceast data into a dataframe.
    """
    # Initialize lists to store forecast
    forecast_list = []

    # Check if more than one day
    if forecast.get("forecasts"):
        forecast_iterator = json_normalize(forecast.get("forecasts"))

    # Single day
    else:
        forecast_iterator = json_normalize(forecast)

    # Loop through each row in the top-level flattened dataframe
    for index, row in forecast_iterator.iterrows():

        # Extract date, lat, lon for insertion into lower-level
        #  dataframe outputs
        date = row["date"]
        lat = row["location.latitude"]
        lon = row["location.longitude"]

        # Check forecast type
        # Main forecast
        if forecast_type == "main":

            # Extract the main forecast from the top-level dataframe
            forecast = row["forecast"]

            # Faltten data into dataframe
            forecast_norm = json_normalize(forecast)

            # Drop soil moisture and soil temperature columns
            #  (will be extracted as indivdiual dataframes)
            forecast_norm.drop(
                columns=["soilTemperatures", "soilMoisture"],
                axis=1,
                inplace=True,
            )
            # Assign date, lat/lon to dataframe
            forecast_norm["date"] = date
            forecast_norm["latitude"] = lat
            forecast_norm["longitude"] = lon

            # Set date as index
            forecast_norm.set_index(["date"], inplace=True)

            # Add the dataframe to a list of dataframes
            forecast_list.append(forecast_norm)

        # Soil forecast
        elif forecast_type == "soil":

            # Get soil temperature data
            forecast_soil_temp = row["forecast"][0].get("soilTemperatures")
            forecast_soil_moisture = row["forecast"][0].get("soilMoisture")

            # Flatten data into dataframe
            forecast_soil_temp_df = json_normalize(forecast_soil_temp)
            forecast_soil_moisture_df = json_normalize(forecast_soil_moisture)

            # Combine temperature and moisture
            forecast_soil_df = forecast_soil_temp_df.merge(
                forecast_soil_moisture_df,
                on="depth",
                suffixes=("_temp", "_moisture"),
            )

            # Assign date, lat/lon to dataframe
            forecast_soil_df["date"] = date
            forecast_soil_df["latitude"] = lat
            forecast_soil_df["longitude"] = lon

            # Shorten depth values to numerics (will be used in MultiIndex)
            forecast_soil_df["depth"] = forecast_soil_df["depth"].apply(
                lambda x: x[0:-15]
            )

            # Rename depth prior to indexing
            forecast_soil_df.rename(
                columns={"depth": "ground_depth_m"}, inplace=True
            )

            # Create multi-index dataframe for date and soil
            #  depth (rename depth columns? rather long)
            soil_multi_index = forecast_soil_df.set_index(
                ["date", "ground_depth_m"]
            )

            # Add dataframe to list of dataframes
            forecast_list.append(soil_multi_index)

        # Invalid forecast type
        else:
            raise ValueError(
                "Invalid forecast type. Must be 'main' or 'soil'."
            )

    # Merge list of dataframes into a single dataframe
    merged = pd.concat(forecast_list)

    # Return merged dataframe
    return merged


def _clean_weather_forecast(df):
    """Helper function that cleans and georeferences weather forecast data.
    """
    # Define global variables
    global FORECAST_COORD_COLS, FORECAST_DROP_COLS, FORECAST_RENAME_MAP

    # Define CRS (EPSG 4326)
    crs = "epsg:4326"

    # Create copy of input dataframe; prevents altering the original
    df_copy = df.copy()

    # Convert to geodataframe
    gdf = gpd.GeoDataFrame(
        df_copy,
        crs=crs,
        geometry=gpd.points_from_xy(
            df[FORECAST_COORD_COLS[0]], df[FORECAST_COORD_COLS[1]]
        ),
    )

    # Add lat/lon columns to drop columns list
    FORECAST_DROP_COLS += FORECAST_COORD_COLS

    # Drop columns
    gdf.drop(columns=FORECAST_DROP_COLS, axis=1, inplace=True, errors="ignore")

    # Rename columns
    gdf.rename(columns=FORECAST_RENAME_MAP, inplace=True, errors="ignore")

    # Return cleaned up geodataframe
    return gdf


def get_weather_forecast(key, secret, forecast_type="main", kwargs=None):
    """Gets weather forecast data from the aWhere API.

    API reference: https://docs.awhere.com/knowledge-base-docs/weather/

    Parameters
    ----------
    key : str
        API key for a valid aWhere API application.

    secret : str
        API secret for a valid aWhere API application.

    forecast_type : str
        Type of weather forecast. Valid options include 'main' and 'soil'.
        Default value is 'main'.

    kwargs : dict, optional
        Keyword arguments for different query parameters.
        Running the function without kwargs will use the
        default values. Arguments include:

            input_type: str
                Type of data, weather API by geolocation or API by field.
                Valid options include 'location' and 'field'. Default
                value is 'location'.

            location: tuple of float
                Tuple containing the location (longitude, latitude) that the
                API gets data from. For use with the 'location' input_type.
                Default value is (-105.648222, 40.313250), for Bear Lake, CO.

            field_id: str
                Field ID for a valid aWhere field associated with the input API
                key/secret. For use with the 'field' input_type. Default value
                is None.

            start_date: str
                Start date for the weather forecast data to be retrieved.
                Formatted 'YYYY-MM-DD'. Valid dates inlclude up to 15 days
                from the current day. Default value is None, which returns the
                next 10 days of forecasted weather from the current day.

            end_date: str
                End date for the weather forecast data to be retrieved.
                Formatted 'YYYY-MM-DD'. Valid dates inlclude up to 15 days
                from the current day. Default value is None, which returns the
                next 10 days of forecasted weather from the current day.

            limit: int
                Number of results in the API response per page. Used with
                offset kwarg to sort through pages of results (cases where the
                number of results exceeds the limit per page). Applicable when
                the number of results exceeds 1. Maximum value is 10. Default
                value is 10.

            offset: int
                Number of results in the API response to skip. Used with limit
                kwarg to sort through pages of results (cases where the
                number of results exceeds the limit per page). Applicable when
                the number of results exceeds 1. Default value is 0 (start
                with first result).

            block_size: int
                Number of hourly forecasts per result. Valid options include
                1, 2, 3, 4, 6, 8, 12, 24. Value of 1 indicates that the API
                returns each daily forecast in 24 1-hour blocks. Value of 6
                indicates that API returns each daily forecast in 4 6-hour
                blocks. Value of 24 indicates that the API returns each daily
                forecast in 1 24-hour block. Default value is 24.

    Returns
    -------
    forecast_gdf : geopandas geodataframe
        Geotdataframe containing the weather forecast data for the
        specified location or aWhere field, for the specified date(s).

    Example
    -------
        >>> # Imports
        >>> import os
        >>> import awherepy as aw
        >>> import awherepy.weather as aww
        >>> # Get aWhere API key and secret
        >>> awhere_api_key = os.environ.get('AWHERE_API_KEY')
        >>> awhere_api_secret = os.environ.get('AWHERE_API_SECRET')
        >>> # Get the weather forecast with default kwargs (Bear Lake, CO)
        >>> bear_lake_forecast = aww.get_weather_forecast(
        ...     key=awhere_api_key, secret=awhere_api_secret)
        >>> # Check number of entries in geodataframe
        >>> len(bear_lake_forecast)
        10
        >>> # Define non-default kwargs (Manchester Center, Vermont)
        >>> # Define kwargs for Manchester, Vermont
        >>> vt_kwargs = {
        ...     'location': (-73.0723269, 43.1636875)
        ... }
        >>> # Get weather forecast for Manchester, Vermont
        >>> manchester_vt_forecast = aww.get_weather_forecast(
        ...     key=awhere_api_key,
        ...     secret=awhere_api_secret,
        ...     kwargs=vt_kwargs
        ... )
        >>> # Check number entries in geodataframe
        >>> len(manchester_vt_forecast)
        10
        >>> # Get soil forecast for default values (Bear Lake, CO)
        >>> bear_lake_forecast_soil = aww.get_weather_forecast(
        ...     key=awhere_api_key,
        ...     secret=awhere_api_secret,
        ...     forecast_type='soil'
        ... )
        >>> # Check number of entries in geodataframe
        >>> len(bear_lake_forecast_soil)
        40
    """
    # Check if credentials are valid
    if aw.valid_credentials(key, secret):

        # Call data
        forecast_json = (
            _call_weather_forecast(key, secret, **kwargs)
            if kwargs
            else _call_weather_forecast(key, secret)
        )

        # Extract data
        forecast_df = _extract_weather_forecast(
            forecast_json, forecast_type=forecast_type
        )

        # Clean data
        forecast_gdf = _clean_weather_forecast(forecast_df)

    else:
        # Raise error
        raise ValueError("Invalid aWhere API credentials.")

    return forecast_gdf



"""
awherepy.models
===============
A module to access and retrieve aWhere agronomic model data.
"""

import requests as rq
import pandas as pd
from pandas.io.json import json_normalize
import geopandas as gpd
import awherepy as aw
import awherepy.fields as awf

# Define valid models
MODELS_LIST = [
    "BarleyGenericMSU",
    "BarleyGenericNDAWN",
    "CanolaBNapusMSU",
    "CanolaBRapaMSU",
    "CanolaGenericNDAWN",
    "Cotton2600UGCE",
    "Cotton2200NCCA",
    "OatGenericMSU",
    "SugarbeetGenericNDAWN",
    "SunflowerEarlyDwarfMSU",
    "SunflowerGenericNDAWN",
    "WheatHardRedMSU",
    "WheatGenericMAWG",
    "WheatGenericNDAWN",
    "WheatGenericOSU",
    "WheatGenericVCE",
    "Corn2300ISUAbendroth",
    "Corn2500ISUAbendroth",
    "Corn2700ISUAbendroth",
    "Corn2900ISUAbendroth",
    "SorghumShortSeasonTexasAM",
    "SorghumLongSeasonTexasAM",
    "MaizeKALRO",
    "SunflowerH8998KALRO",
    "SunflowerFedhaKALRO",
    "GreenGramKALRO",
    "PotatoKALRO",
    "SorghumKALRO",
    "SoyaKALRO",
]

# Define variables for data cleaning
# Models
MODELS_DROP_COLS = [
    "_links.self.href",
    "_links.curies",
    "_links.awhere:crop",
    "_links.awhere:modelDetails.href",
]

MODELS_RENAME_MAP = {
    "id": "model_id",
    "name": "model_name",
    "description": "model_description",
    "type": "model_type",
    "source.name": "model_source",
    "source.link": "model_link",
}


# Define variables for data cleaning
MODEL_DETAILS_DROP_COLS = [
    "gddUnits",
    "stages",
    "_links.self.href",
    "_links.curies",
    "_links.awhere:model.href",
]

MODEL_DETAILS_BASE_RENAME_MAP = {
    "biofix": "biofix_days",
    "gddMethod": "gdd_method",
    "gddBaseTemp": "gdd_base_temp_cels",
    "gddMaxBoundary": "gdd_max_boundary_cels",
    "gddMinBoundary": "gdd_min_boundary_cels",
}

MODEL_DETAILS_STAGE_RENAME_MAP = {
    "id": "stage_id",
    "stage": "stage_name",
    "description": "stage_description",
    "gddThreshold": "gdd_threshold_cels",
}

# Model results
MODEL_RESULTS_DROP_COLS = ["longitude", "latitude"]

MODEL_RESULTS_RENAME_MAP = {
    "date": "stage_start_date",
    "id": "stage_id",
    "stage": "stage_name",
    "description": "stage_description",
    "gddThreshold": "gdd_threshold_cels",
    "accumulatedGdds": "gdd_accumulation_current_cels",
    "gddRemaining": "gdd_remaining_next_cels",
}

MODEL_RESULTS_REORDER_COLS = [
    "model_id",
    "biofix_date",
    "planting_date",
    "stage_start_date",
    "stage_id",
    "stage_name",
    "stage_description",
    "gdd_threshold_cels",
    "gdd_accumulation_current_cels",
    "gdd_remaining_next_cels",
    "geometry",
]


def get_models(key, secret, model_id=None):
    """Retrieve a list of available models or single model if specified.

    API reference: https://docs.awhere.com/knowledge-base-docs/models/

    Parameters
    ----------
    key : str
        API key for a valid aWhere API application.

    secret : str
        API secret for a valid aWhere API application.

    model_id : str, optional
        Valid aWhere model id.

    Returns
    -------
    model_df : pandas dataframe
        Dataframe containing a single model or all available models.

    Example
    -------
        >>> # Imports
        >>> import os
        >>> import awherepy as aw
        >>> import awherepy.models as awm
        >>> # Get aWhere API key and secret
        >>> awhere_api_key = os.environ.get('AWHERE_API_KEY')
        >>> awhere_api_secret = os.environ.get('AWHERE_API_SECRET')
        >>> # Get all available aWhere models
        >>> all_models = awm.get_models(awhere_api_key, awhere_api_secret)
        >>> # Display model id for first model
        >>> all_models.index[0]
        BarleyGenericMSU
        >>> # Get a single model
        >>> corn_2700 = awm.get_models(
        ...     awhere_api_key,
        ...     awhere_api_secret,
        ...     modei_id='Corn2700ISUAbendroth'
        ... )
        >>> # Show model name
        >>> corn_2700.model_name
        model_id
        Corn2700ISUAbendroth    Corn, 2700 GDD (Iowa State University)
        Name: model_name, dtype: object
        >>> # Show model type
        >>> corn_2700.model_type
        model_id
        Corn2700ISUAbendroth    GrowthStage
        Name: model_type, dtype: object
    """
    # Define global variables
    global MODELS_LIST, MODELS_DROP_COLS, MODELS_RENAME_MAP

    # Check if credentials are valid
    if aw.valid_credentials(key, secret):

        # Define models api url
        api_url = "https://api.awhere.com/v2/agronomics/models"

        # Get OAuth token
        auth_token = aw.get_oauth_token(key, secret)

        # Set up the HTTP request headers
        auth_headers = {"Authorization": f"Bearer {auth_token}"}

        # Check if single model
        if model_id:

            # Raise error if model id is invalid
            if model_id not in MODELS_LIST:
                raise KeyError(
                    (
                        f"Invalid model."
                        f"Valid models: {', '.join(MODELS_LIST)}"
                    )
                )

            # Get API response
            response = rq.get(f"{api_url}/{model_id}", headers=auth_headers)

            # Convert API response to JSON format
            response_json = response.json()

            # Convert to dataframe
            model_df = json_normalize(response_json)

        # All crops
        else:

            # Initialize variables
            offset = 0
            limit = 10
            pages = 3

            # Initialize list to store dataframe for each page of results
            response_df_list = []

            # Loop through all pages
            while offset < limit * pages:

                # Get API response
                response = rq.get(
                    f"{api_url}?limit={limit}&offset={offset}",
                    headers=auth_headers,
                )

                # Convert API response to JSON format
                response_json = response.json()

                # Convert json to dataframe
                response_df_loop = json_normalize(response_json.get("models"))

                # Append to dataframe list
                response_df_list.append(response_df_loop)

                # Increase offset by 1 page of results
                offset += limit

            # Merge all page dataframes into a single dataframe
            model_df = pd.concat(response_df_list, axis=0)

        # Drop unnecessary columns
        model_df.drop(
            columns=MODELS_DROP_COLS, inplace=True,
        )

        # Reset index
        model_df.reset_index(drop=True, inplace=True)

        # Rename columns
        model_df.rename(columns=MODELS_RENAME_MAP, inplace=True)

        # Set index to model id
        model_df.set_index(keys=["model_id"], drop=True, inplace=True)

    # Invalid credentials
    else:
        # Raise error
        raise ValueError("Invalid aWhere API credentials.")

    # Return models
    return model_df


def get_model_details(key, secret, model_id=None):
    """Retrieve model details for all available models
    or single model if specified.

    API reference: https://docs.awhere.com/knowledge-base-docs/models/

    Parameters
    ----------
    key : str
        API key for a valid aWhere API application.

    secret : str
        API secret for a valid aWhere API application.

    model_id : str, optional
        Valid aWhere model id.

    Returns
    -------
    tuple

        base_info_df: pandas dataframe
            Dataframe containing the base information about the model.

        stage_info_df: pandas dataframe
            Dataframe containing the stage information about the model.

    Example
    -------
        >>> # Imports
        >>> import os
        >>> import awherepy as aw
        >>> import awherepy.models as awm
        >>> # Get aWhere API key and secret
        >>> awhere_api_key = os.environ.get('AWHERE_API_KEY')
        >>> awhere_api_secret = os.environ.get('AWHERE_API_SECRET')
        >>> # Get details for barley generic
        >>> barley_base_details, barley_stage_details = awm.get_model_details(
        ...     awhere_api_key, awhere_api_secret, model_id='BarleyGenericMSU'
        ... )
        >>> # Show base detiils - GDD method
        >>> barley_base_details.gdd_method
        model_id
        BarleyGenericMSU    standard
        Name: gdd_method, dtype: object
        >>> # Show stage names
        >>> barley_stage_details.stage_name
        model_id          stage_id
        BarleyGenericMSU  stage1      Universal 1.0
                          stage2      Universal 1.2
                          stage3      Universal 2.1
                          stage4      Universal 3.1
                          stage5      Universal 6.1
                          stage6      Universal 7.1
                          stage7      Universal 8.5
                          stage8      Universal 8.9
        Name: stage_name, dtype: object
    """
    # Define global variables
    global MODELS_LIST, MODELS_DROP_COLS, MODELS_RENAME_MAP

    # Check if credentials are valid
    if aw.valid_credentials(key, secret):

        # Define models api url
        api_url = "https://api.awhere.com/v2/agronomics/models"

        # Get OAuth token
        auth_token = aw.get_oauth_token(key, secret)

        # Set up the HTTP request headers
        auth_headers = {"Authorization": f"Bearer {auth_token}"}

        # Check if single model
        if model_id:

            # Get API response
            response = rq.get(
                f"{api_url}/{model_id}/details", headers=auth_headers
            )

            # Convert API response to JSON format
            response_json = response.json()

            # Get model base information
            base_info_df = json_normalize(response_json)

            # Drop columns
            base_info_df.drop(
                columns=MODEL_DETAILS_DROP_COLS, inplace=True, errors="ignore"
            )

            # Add model id
            base_info_df["model_id"] = model_id

            # Set index to model id
            base_info_df.set_index("model_id", inplace=True)

            # Rename columns
            base_info_df.rename(
                columns=MODEL_DETAILS_BASE_RENAME_MAP,
                inplace=True,
                errors="ignore",
            )

            # Get model stage information
            stage_info_df = json_normalize(response_json.get("stages"))

            # Drop columns
            stage_info_df.drop(
                columns=MODEL_DETAILS_DROP_COLS, inplace=True, errors="ignore"
            )

            # Add model id
            stage_info_df["model_id"] = model_id

            # Rename columns
            stage_info_df.rename(
                columns=MODEL_DETAILS_STAGE_RENAME_MAP,
                inplace=True,
                errors="ignore",
            )

            # Set multiindex to model id and stage id
            stage_info_df.set_index(["model_id", "stage_id"], inplace=True)

        # All models
        else:

            # Initialize lists to store dataframes
            base_list = []
            stage_list = []

            # Loop through all available models
            for model in list(get_models(key, secret).index):

                # Get API response
                response = rq.get(
                    f"{api_url}/{model}/details", headers=auth_headers
                )

                # Convert API response to JSON format
                response_json = response.json()

                # Get model base information
                base = json_normalize(response_json)

                # Drop columns
                base.drop(
                    columns=MODEL_DETAILS_DROP_COLS,
                    inplace=True,
                    errors="ignore",
                )

                # Add model id
                base["model_id"] = model

                # Set index to model id
                base.set_index("model_id", inplace=True)

                # Rename columns
                base.rename(
                    columns=MODEL_DETAILS_BASE_RENAME_MAP,
                    inplace=True,
                    errors="ignore",
                )

                # Append individual base dataframes to list
                base_list.append(base)

                # Get model stage information
                stage = json_normalize(response_json.get("stages"))

                # Drop columns
                stage.drop(
                    columns=MODEL_DETAILS_DROP_COLS,
                    inplace=True,
                    errors="ignore",
                )

                # Add model id
                stage["model_id"] = model

                # Rename columns
                stage.rename(
                    columns=MODEL_DETAILS_STAGE_RENAME_MAP,
                    inplace=True,
                    errors="ignore",
                )

                # Set multiindex to model id and stage id
                stage.set_index(["model_id", "stage_id"], inplace=True)

                # Append individual stage dataframes to list
                stage_list.append(stage)

            # Merge base and stage dataframes
            base_info_df = pd.concat(base_list, axis=0)
            stage_info_df = pd.concat(stage_list, axis=0)

    # Invalid credentials
    else:
        # Raise error
        raise ValueError("Invalid aWhere API credentials.")

    # Return base info and stage info dataframes
    return base_info_df, stage_info_df


def get_model_results(key, secret, field_id, model_id):
    """Returns the aWhere agronomic model results
    associated with a specific field and model.

    API reference: https://docs.awhere.com/knowledge-base-docs/models/

    Parameters
    ----------
    key : str
        API key for a valid aWhere API application.

    secret : str
        API secret for a valid aWhere API application.

    field_id : str
        Field id for an existing aWhere field.

    model_id : str
        Valid aWhere model id.

    Returns
    -------
    stages_df: pandas dataframe
        Dataframe containing the model results.

    Example
    -------
        >>> # Imports
        >>> import os
        >>> import awherepy as aw
        >>> import ahwerepy.plantings as awp
        >>> import awherepy.models as awm
        >>> # Get aWhere API key and secret
        >>> awhere_api_key = os.environ.get('AWHERE_API_KEY')
        >>> awhere_api_secret = os.environ.get('AWHERE_API_SECRET')
        >>> # Create planting for Manchester, Vermont field
        >>> vt_planting_info = {
        ...     'crop': 'wheat-hardred',
        ...     'planting_date': '2020-06-01'
        ... }
        >>> awp.create_planting(
        ...     awhere_api_key,
        ...     awhere_api_secret,
        ...     field_id='VT-Manchester',
        ...     planting_info=vt_info
        ... )
        >>> # Get model results
        >>> vt_model_results = awm.get_model_results(
        ...     awhere_api_key,
        ...     awhere_api_secret,
        ...     field_id='VT-Manchester',
        ...     model_id='WheatHardRedMSU'
        ... )
        >>> # Show some columns of model results
        >>> vt_model_results[['stage_description', 'gdd_threshold_cels']]

                                   stage_description  gdd_threshold_cels
        field_id      stage_status
        VT-Manchester Current              emergence               161.0
                      Next          leaf development               206.0
    """
    # Define global variables
    global MODEL_RESULTS_DROP_COLS
    global MODEL_RESULTS_RENAME_MAP
    global MODEL_RESULTS_REORDER_COLS

    # Check if credentials are valid
    if aw.valid_credentials(key, secret):

        # Raise error if field does not exist
        if field_id not in awf.get_fields(key, secret).index:
            raise KeyError("Field does not exist within account.")

        # Define api url
        api_url = (
            f"https://api.awhere.com/v2/agronomics/fields/"
            f"{field_id}/models/{model_id}/results"
        )

        # Get OAuth token
        auth_token = aw.get_oauth_token(key, secret)

        # Setup the HTTP request headers
        auth_headers = {
            "Authorization": f"Bearer {auth_token}",
        }

        # Get API response
        response = rq.get(f"{api_url}", headers=auth_headers)

        # Convert API response to JSON format
        response_json = response.json()

        # Get stage info
        # Previous stage
        try:
            # Get info
            previous_stage_df = json_normalize(
                response_json.get("previousStages")
            )

            # Add column
            previous_stage_df["stage_status"] = "Previous"

        # Catch error if no previous stage
        except TypeError:
            previous_stage_df = pd.DataFrame(
                columns=["stage_status"], data=["Previous"]
            )

        # Current stage
        try:
            # Get info
            current_stage_df = json_normalize(
                response_json.get("currentStage")
            )

            # Add column
            current_stage_df["stage_status"] = "Current"

        # Catch error if no current stage
        except TypeError:
            current_stage_df = pd.DataFrame(
                columns=["stage_status"], data=["Current"]
            )

        # Next stage
        try:
            # Get info
            next_stage_df = json_normalize(response_json.get("nextStage"))

            # Add column
            next_stage_df["stage_status"] = "Next"

        # Catch error if no next stage
        except TypeError:
            next_stage_df = pd.DataFrame(
                columns=["stage_status"], data=["Next"]
            )

        # Merge into one dataframe
        stages_df = pd.concat(
            [previous_stage_df, current_stage_df, next_stage_df],
            sort=False,
            axis=0,
        )

        # Change column names
        stages_df.rename(
            columns=MODEL_RESULTS_RENAME_MAP, inplace=True, errors="ignore",
        )

        # Add base data
        stages_df["biofix_date"] = response_json.get("biofixDate")
        stages_df["planting_date"] = response_json.get("plantingDate")
        stages_df["model_id"] = response_json.get("modelId")
        stages_df["field_id"] = response_json.get("location").get("fieldId")
        stages_df["longitude"] = response_json.get("location").get("longitude")
        stages_df["latitude"] = response_json.get("location").get("latitude")

        # Set index
        stages_df.set_index(["field_id", "stage_status"], inplace=True)

        # Prep for geodataframe conversion
        df_copy = stages_df.copy()

        # Define CRS (EPSG 4326)
        crs = "epsg:4326"

        # Convert to geodataframe
        stages_gdf = gpd.GeoDataFrame(
            df_copy,
            crs=crs,
            geometry=gpd.points_from_xy(df_copy.longitude, df_copy.latitude),
        )

        # Drop lat/lon
        stages_gdf.drop(
            columns=MODEL_RESULTS_DROP_COLS, inplace=True, errors="ignore"
        )

        # Reorder columns
        stages_gdf = stages_gdf.reindex(columns=MODEL_RESULTS_REORDER_COLS)

    # Invalid credentials
    else:
        # Raise error
        raise ValueError("Invalid aWhere API credentials.")

    # Return model geodataframe
    return stages_gdf



"""
awherepy.fields
===============
A module to access and work with aWhere fields.
"""

import requests as rq
from pandas.io.json import json_normalize
import geopandas as gpd
import awherepy as aw

# Define lon/lat columns
FIELD_COORD_COLS = ["centerPoint.longitude", "centerPoint.latitude"]


# Define columns to drop
FIELD_DROP_COLS = [
    "_links.self.href",
    "_links.curies",
    "_links.awhere:observations.href",
    "_links.awhere:forecasts.href",
    "_links.awhere:plantings.href",
    "_links.awhere:agronomics.href",
]

# Define new column names
FIELD_RENAME_MAP = {
    "name": "field_name",
    "acres": "area_acres",
    "farmId": "farm_id",
    "id": "field_id",
}

# Define CRS (EPSG 4326)
FIELD_CRS = "epsg:4326"


def get_fields(key, secret, kwargs=None):
    """
    Returns fields (all or individual) associated with an aWhere application.

    API reference: https://docs.awhere.com/knowledge-base-docs/field-plantings/

    Parameters
    ----------
    key : str
        API key for a valid aWhere API application.

    secret : str
        API secret for a valid aWhere API application.

    kwargs : dict, optional
        Keyword arguments for different query parameters.
        Running the function without kwargs will use the
        default values. Arguments include:

            field_id: str
                Field ID for a valid aWhere field associated with the input API
                key/secret. Causes the function to return a single field if the
                value is not None. If value is None, all fields associated with
                the API key/secret will be returned. Default value is None.

            limit: int
                Number of results in the API response per page. Used with
                offset kwarg to sort through pages of results (cases where the
                number of results exceeds the limit per page). Applicable when
                the number of results exceeds 1. Maximum value is 10. Default
                value is 10.

            offset: int
                Number of results in the API response to skip. Used with limit
                kwarg to sorts through pages of results (cases where the
                number of results exceeds the limit per page). Applicable when
                the number of results exceeds 1. Default value is 0 (start
                with first result).

    Returns
    -------
    fields_gdf : geopandas geodataframe
        Geodataframe containing all fields or an individual field.

    Example
    -------
        >>> # Imports
        >>> import os
        >>> import awherepy as aw
        >>> import awherepy.fields as awf
        >>> # Get aWhere API key and secret
        >>> awhere_api_key = os.environ.get('AWHERE_API_KEY')
        >>> awhere_api_secret = os.environ.get('AWHERE_API_SECRET')
        >>> # Get all fields associated with key/secret
        >>> all_fields = awf.get_fields(awhere_api_key, awhere_api_secret)
        >>> # Check number of entries in field
        >>> len(all_fields)
        3
        >>> # Define kwargs for single field
        >>> manchester_vt_kwargs = {'field_id': 'Manchester-VT'}
        >>> # Get a single field
        >>> manchester_vt_field = awf.get_fields(
        ...     awhere_api_key, awhere_api_secret, kwargs=manchester_vt_kwargs)
        >>> # Check number of entries
        >>> len(manchester_vt_field)
        1
    """
    # Define global variables
    global FIELD_COORD_COLS, FIELD_DROP_COLS, FIELD_RENAME_MAP, FIELD_CRS

    # Check if credentials are valid
    if aw.valid_credentials(key, secret):

        # Check if kwargs is not defined
        if not kwargs:

            # Set limit and offset query parameters
            kwargs = {"limit": 10, "offset": 0}

        # Define fields api url
        api_url = "https://api.awhere.com/v2/fields"

        # Get OAuth token
        auth_token = aw.get_oauth_token(key, secret)

        # Set up the HTTP request headers
        auth_headers = {"Authorization": f"Bearer {auth_token}"}

        # Single field
        if kwargs.get("field_id"):
            # Perform the HTTP request to obtain a specific field
            fields_response = rq.get(
                f"{api_url}/{kwargs.get('field_id')}", headers=auth_headers
            )

        # All fields
        else:
            # Perform the HTTP request to obtain a list of all fields
            fields_response = rq.get(
                (
                    f"{api_url}?limit={kwargs.get('limit')}"
                    f"&offset={kwargs.get('offset')}"
                ),
                headers=auth_headers,
            )

        # Convert API response to JSON
        response = fields_response.json()

        # Convert JSON to dataframe
        fields_df = (
            json_normalize(response)
            if kwargs.get("field_id")
            else json_normalize(response.get("fields"))
        )

        # Prep for geodataframe conversion
        df_copy = fields_df.copy()

        # Convert to geodataframe
        fields_gdf = gpd.GeoDataFrame(
            df_copy,
            crs=FIELD_CRS,
            geometry=gpd.points_from_xy(
                df_copy[FIELD_COORD_COLS[0]], df_copy[FIELD_COORD_COLS[1]]
            ),
        )

        # Add lat/lon columns to drop columns list
        FIELD_DROP_COLS += FIELD_COORD_COLS

        # Drop columns
        fields_gdf.drop(columns=FIELD_DROP_COLS, inplace=True, errors="ignore")

        # Rename columns
        fields_gdf.rename(
            columns=FIELD_RENAME_MAP, inplace=True, errors="ignore"
        )

        # Set field ID as index
        fields_gdf.set_index("field_id", inplace=True)

    # Invalid credentials
    else:
        # Raise error
        raise ValueError("Invalid aWhere API credentials.")

    # Return geodataframe
    return fields_gdf


def create_field(key, secret, field_info):
    """Creates a field associated with an aWhere application.

    API reference: https://docs.awhere.com/knowledge-base-docs/field-plantings/

    Parameters
    ----------
    key : str
        API key for a valid aWhere API application.

    secret : str
        API secret for a valid aWhere API application.

    field_info : dict
        Dictionary containing the information required to create an aWhere
        field. Can contain the following keys:

            field_id: str, required
                ID for the field. This ID is used to reference the field in
                all applicable aWhere APIs. Can contain only alphanumeric
                characters, underscores, and dashes. Maximum 50 characters.

            center_latitude: int or float, required
                Latitiude (in decimal degrees) of the field center point.

            center_longitude: int or float, required
                Longitude (in decimal degrees) of the field center point.

            farm_id: str, required
                Name of the farm. Not referenced by any other aWhere APIs. Can
                contain only alphanumeric characters, underscores, and dashes.
                Maximum 50 characters.

            field_name: str, optional
                Name of the field. Not referenced by any other aWhere APIs.
                Can contain only alphanumeric characters, underscores, and
                dashes. Maximum 255 characters.

            acres: int or float, optional
                Size of the field in acres.

    Returns
    -------
    field : geopandas geodataframe
        Geodataframe containing the newly-created aWhere field.

    Example
    -------
        >>> # Imports
        >>> import os
        >>> import awherepy as aw
        >>> import awherepy.fields as awf
        >>> # Get aWhere API key and secret
        >>> awhere_api_key = os.environ.get('AWHERE_API_KEY')
        >>> awhere_api_secret = os.environ.get('AWHERE_API_SECRET')
        >>> # Define field information for Manchester, Vermont
        >>> manchester_field_info = {
        ...     'field_id': 'VT-Manchester',
        ...     'field_name': 'Manchester-Field',
        ...     'farm_id': 'Manchester-Farm',
        ...     'center_latitude': 43.1636875,
        ...     'center_longitude': -73.0723269,
        ...     'acres': 1
        ... }
        >>> Create field
        >>> field = awf.create_field(
        ...    api_key, api_secret, field_info=manchester_field_info
        ... )
        Attempting to create field...
        Created field: VT-Manchester
        >>> # Check geodataframe index (field id)
        >>> print(field.index[0])
        VT-Manchester
        >>> # Display field
        >>> field
    """
    # Check if credentials are valid
    if aw.valid_credentials(key, secret):

        # Check field_info object type
        if not isinstance(field_info, dict):
            raise TypeError(
                "Invalid type: 'field_info' must be of type dictionary."
            )

        # Raise errors for missing required parameters
        if not field_info.get("field_id"):
            raise KeyError("Missing required field parameter: 'field_id'.")

        if not field_info.get("center_latitude"):
            raise KeyError(
                ("Missing required field parameter: 'center_latitude'.")
            )

        if not field_info.get("center_longitude"):
            raise KeyError(
                ("Missing required field parameter: 'center_longitude'.")
            )

        # Raise error if field name already exists
        if field_info.get("field_id") in get_fields(key, secret).index:
            raise KeyError("Field name already exists within account.")

        # Define fields api url
        api_url = "https://api.awhere.com/v2/fields"

        # Get OAuth token
        auth_token = aw.get_oauth_token(key, secret)

        # Setup the HTTP request headers
        auth_headers = {
            "Authorization": f"Bearer {str(auth_token)}",
            "Content-Type": "application/json",
        }

        # Define request body
        field_body = {
            "id": field_info.get("field_id"),
            "name": field_info.get("field_name"),
            "farmId": field_info.get("farm_id"),
            "centerPoint": {
                "latitude": field_info.get("center_latitude"),
                "longitude": field_info.get("center_longitude"),
            },
            "acres": field_info.get("acres"),
        }

        # Perform the POST request to create Field
        print("Attempting to create field...")
        response = rq.post(api_url, headers=auth_headers, json=field_body)

        # Check if request succeeded
        if response.ok:

            # Get field for output
            field = get_fields(
                key, secret, kwargs={"field_id": field_info.get("field_id")}
            )

            # Indicate success
            print(f"Created field: {field_info.get('field_id')}")

        else:

            # Indicate error
            print("Failed to create field.")

    # Invalid credentials
    else:
        # Raise error
        raise ValueError("Invalid aWhere API credentials.")

    # Return created field
    return field


def update_field(key, secret, field_info):
    """Update the field name and/or farm id for a specified field.

    API reference: https://docs.awhere.com/knowledge-base-docs/field-plantings/

    Parameters
    ----------
    key : str
        API key for a valid aWhere API application.

    secret : str
        API secret for a valid aWhere API application.

    field_info : dict
        Dictionary containing information required to update the field. Must
        provide an update to field_name or farm_id (cannot make empty update).
        Can contain the following keys:

            field_id: str, required
                Field ID for an existing aWhere field. This identifies the
                field that will be updated.

            field_name: str, required if farm_id not specified
                Updated field name.

            farm_id: str, required if field_name is not specified
                Updated farm ID.

    Returns
    -------
    field : geopandas geodataframe
        Geodataframe containing the updated aWhere field.

    Example
    -------
        >>> # Imports
        >>> import os
        >>> import awherepy as aw
        >>> import awherepy.fields as awf
        >>> # Get aWhere API key and secret
        >>> awhere_api_key = os.environ.get('AWHERE_API_KEY')
        >>> awhere_api_secret = os.environ.get('AWHERE_API_SECRET')
        >>> # Define update parameters for Manchester, Vermont
        >>> manchester_update_info = {
        ...     'field_id': 'VT-Manchester',
        ...     'field_name': 'Manchester-Field-Update',
        ...     'farm_id': 'Manchester-Farm-Update'
        ... }
        >>> # Update the field
        >>> field_updated = awf.update_field(
        ...     api_key, api_secret,
        ...     field_info=manchester_update_info
        ... )
        Attempting to update field...
        Updated field: VT-Manchester
        >>> # Display updated field
        >>> field_updated
    """
    # Check if credentials are valid
    if aw.valid_credentials(key, secret):

        # Check field_info object type
        if not isinstance(field_info, dict):
            raise TypeError(
                "Invalid type: 'field_info' must be of type dictionary."
            )

        # Raise errors for missing required parameters
        if not field_info.get("field_id"):
            raise KeyError("Missing required field parameter: 'field_id'.")

        if not (field_info.get("field_name") or field_info.get("farm_id")):
            raise KeyError(
                (
                    "Missing parameter: must update at least one attribute,"
                    "'field_name' or 'farm_id'."
                )
            )

        # Raise error if field does not exist
        if field_info.get("field_id") not in get_fields(key, secret).index:
            raise KeyError("Field name does not exist within account.")

        # Define fields api url
        api_url = "https://api.awhere.com/v2/fields"

        # Get OAuth token
        auth_token = aw.get_oauth_token(key, secret)

        # Setup the HTTP request headers
        auth_headers = {
            "Authorization": f"Bearer {auth_token}",
        }

        # Set up request body
        # Field name and farm id update
        if field_info.get("field_name") and field_info.get("farm_id"):
            field_body = [
                {
                    "op": "replace",
                    "path": "/name",
                    "value": field_info.get("field_name"),
                },
                {
                    "op": "replace",
                    "path": "/farmId",
                    "value": field_info.get("farm_id"),
                },
            ]

        # Field name only update
        elif field_info.get("field_name") and not field_info.get("farm_id"):
            field_body = [
                {
                    "op": "replace",
                    "path": "/name",
                    "value": field_info.get("field_name"),
                }
            ]

        # Farm id only update
        elif field_info.get("farm_id") and not field_info.get("field_name"):
            field_body = [
                {
                    "op": "replace",
                    "path": "/farmId",
                    "value": field_info.get("farm_id"),
                }
            ]

        # Perform the HTTP request to update field information
        print("Attempting to update field...")
        response = rq.patch(
            f"{api_url}/{field_info.get('field_id')}",
            headers=auth_headers,
            json=field_body,
        )

        # Check if request succeeded
        if response.ok:

            # Get field for output
            field = get_fields(
                key, secret, kwargs={"field_id": field_info.get("field_id")}
            )

            # Indicate success
            print(f"Updated field: {field_info.get('field_id')}")

        else:
            # Indicate error
            field = print("Failed to update field.")

    # Invalid credentials
    else:
        # Raise error
        raise ValueError("Invalid aWhere API credentials.")

    # Return updated field
    return field


def delete_field(key, secret, field_id):
    """
    Deletes a speficied aWhere field.

    API reference: https://docs.awhere.com/knowledge-base-docs/field-plantings/

    Parameters
    ----------
    key : str
        API key for a valid aWhere API application.

    secret : str
        API secret for a valid aWhere API application.

    field_id : str
        Field ID for an existing aWhere field. This identifies the field that
        will be deleted from the aWhere application.

    Returns
    -------
    message : str
        Output message indicating success or failure for the field deletion.

    Example
    -------
    >>> # Imports
    >>> import os
    >>> import awherepy as aw
    >>> import awherepy.fields as awf
    >>> # Get aWhere API key and secret
    >>> awhere_api_key = os.environ.get('AWHERE_API_KEY')
    >>> awhere_api_secret = os.environ.get('AWHERE_API_SECRET')
    >>> # Delete field for Manchester, Vermont
    >>> awf.delete_field(
    ...     awhere_api_key, awhere_api_secret, field_id='VT-Manchester'
    ... )
    Attempting to delete field...
    Deleted field: VT-Manchester
    """
    # Check if credentials are valid
    if aw.valid_credentials(key, secret):

        # Raise error if field does not exist
        if field_id not in get_fields(key, secret).index:
            raise KeyError("Field name does not exist within account.")

        # Define fields api url
        api_url = "https://api.awhere.com/v2/fields"

        # Get OAuth token
        auth_token = aw.get_oauth_token(key, secret)

        # Setup the HTTP request headers
        auth_headers = {
            "Authorization": f"Bearer {auth_token}",
        }

        # Perform the POST request to Ddlete the field
        print("Attempting to delete field...")
        rq.delete(f"{api_url}/{field_id}", headers=auth_headers)

        # Check if field exists within account
        try:
            get_fields(key, secret, kwargs={"field_id": field_id})

        # Catch error if field does not exist (was deleted)
        except KeyError:
            message = f"Deleted field: {field_id}"
            print(message)

        # If delete did not work
        else:
            message = "Could not delete field."
            print(message)

    # Invalid credentials
    else:
        # Raise error
        raise ValueError("Invalid aWhere API credentials.")

    return message



"""
awherepy.grids
==============
A module to create and work with aWhere grids.
"""

import os
import matplotlib.pyplot as plt
import numpy as np
import rasterio as rio
from shapely.geometry import Polygon
import geopandas as gpd
import rasterstats as rs


def create_grid(study_area_path, buffer_distance, cell_size=0.08):
    """Creates an aWhere-sized grid (0.08 x 0.08 degree,
    5 arc-minute x 5 arc-minute grid) fit to a polygon.

    Parameters
    ----------
    study_area_path : str
        Path the polygon shapefile boundary.

    buffer_distance : int or float
        Buffer size in degrees (WGS 84 CRS).

    cell_size: int or float
        Grid size (x and y dimension) in degrees (WGS 84 CRS).

    Returns
    -------
    study_area_grid_4326: geopandas GeoDataFrame
        Grid dataframe shaped to the polygon boundary.

    study_area_4326: geopandas GeoDataFrame
        Study area boundary projected to WGS 84, EPSG 4326.

    Example
    -------
        >>> # Import packages
        >>> import os
        >>> import awherepy.grids as awg
        >>> # Define path to shapefile boundary
        >>> vt_bound_path = os.path.join(
        ...     working_directory, 'shapefiles', 'vermont_state_boundary.shp')
        >>> # Create aWhere grid
        >>> vt_grid, vt_bound_4326 = awg.create_grid(
        ...     vt_bound_path, buffer_distance=0.12)
        >>> # Plot aWhere grid
        >>> vt_grid.plot(facecolor="none", edgecolor="#984ea3", linewidth=1.5)
    """
    # Read shapefile into geodataframe
    study_area = gpd.read_file(study_area_path)

    # Project to WGS 84 Lat/Lon, EPSG 4326 if no CRS match
    if not study_area.crs == "epsg:4326":
        study_area_4326 = study_area.to_crs("epsg:4326")

    else:
        study_area_4326 = study_area

    # Create buffer around state boundary
    study_area_4326_buffer = study_area_4326.buffer(distance=buffer_distance)

    # Convert buffer from geoseries to geodataframe
    study_area_4326_buffer_gdf = gpd.GeoDataFrame(
        study_area_4326_buffer, crs="epsg:4326"
    )

    # Rename geometry column in buffer
    study_area_4326_buffer_gdf.rename(columns={0: "geometry"}, inplace=True)

    # Get extent of buffered boundary
    longitude_min = study_area_4326_buffer_gdf.bounds.minx[0]
    latitude_min = study_area_4326_buffer_gdf.bounds.miny[0]
    longitude_max = study_area_4326_buffer_gdf.bounds.maxx[0]
    latitude_max = study_area_4326_buffer_gdf.bounds.maxy[0]

    # Create lists for lat/lon extents
    longitude_vals = np.arange(longitude_min, longitude_max, cell_size)
    latitude_vals = np.arange(latitude_min, latitude_max, cell_size)

    # Create grid of polygons based on longitude and latitude ranges
    grid_polys_list = [
        Polygon(
            [
                (longitude, latitude),
                (longitude + cell_size, latitude),
                (longitude + cell_size, latitude + cell_size),
                (longitude, latitude + cell_size),
            ]
        )
        for longitude in longitude_vals
        for latitude in latitude_vals
    ]

    # Create geodataframe from grid polygons
    grid_polys_gdf = gpd.GeoDataFrame(crs=4326, geometry=grid_polys_list)

    # Narrow grid cells to those within the buffered boundary
    study_area_grid_4326 = gpd.sjoin(
        grid_polys_gdf, study_area_4326_buffer_gdf, op="within"
    )

    # Drop unnecessary colum
    study_area_grid_4326.drop(columns="index_right", inplace=True)

    # Return gridded geodataframe
    return study_area_grid_4326, study_area_4326


def extract_centroids(grid):
    """Extracts the longitude and latitude centroids
    from a grid of polygons.

    Parameters
    ----------
    grid : geopandas GeoDataFrame
        Grid dataframe with polygon geometry.

    Returns
    -------
    centroid_list: list (of tuples)
        List containing (longitude, latitude) tuples.

    Example
    -------
        >>> # Extract centroids
        >>> vt_grid_centroids = awg.extract_centroids(vt_grid)
        >>> # Show number of centroids/grid cells
        >>> len(vt_grid_centroids)
        533
        >>> # Show first centroid
        >>> vt_grid_centroids[0]
        (-73.43784136769847, 43.527012318617274)
    """
    # Create copy of dataframe (avoids altering the original)
    grid_extract = grid.copy()

    # Add centroid to each grid cell
    grid_extract["centroid"] = grid_extract.geometry.apply(
        lambda poly: poly.centroid
    )

    # Extract latitude and longitude to new columns
    grid_extract["longitude"] = grid_extract.centroid.apply(
        lambda point: point.x
    )

    grid_extract["latitude"] = grid_extract.centroid.apply(
        lambda point: point.y
    )

    # Extract centroid (as tuples) from grid
    centroid_list = [
        (row.longitude, row.latitude) for row in grid_extract.itertuples()
    ]

    # Return centroids
    return centroid_list


def rasterize(awhere_grid, raster_path, zonal_stats="count sum"):
    """Rasterizes values from a GeoTiff (or other
    georeferenced raster) to the aWhere grid (9x9 km).

    Aggregates the sum by default.

    Parameters
    ----------
    awhere_grid : geopandas geodataframe
        Geodataframe containing a grid of aWhere-sized
        cells, sized to fit a shapefile boundary.

    raster_path : str
        Path to the data the will be rasterized.

    zonal_stats : space-delimited str (optional)
        Zonal statistics to calculate. Default value
        is 'count sum'.

    Returns
    -------
    awhere_grid_rasterized : geopandas geodataframe
        Input geodataframe with the rasterized values
        and grid centroids added.

    Example
    -------
        >>> # Import packages
        >>> import os
        >>> import awherepy.grids as awg
        >>> # Define path to shapefile boundary
        >>> vt_bound_path = os.path.join(
        ...     working_directory, 'shapefiles',
        ...     vermont_state_boundary.shp')
        >>> # Create aWhere grid for Vermont
        >>> vt_grid, vt_bound_4326 = awg.create_grid(
        ...     vt_bound_path, buffer_distance=0.12)
        >>> # Define path to Vermont 2020 population per pixel
        >>> vt_pop_path = os.path.join(
        ...     working_directory, 'geotiffs',
        ...     'vt_ppp_2020.tif')
        >>> # Rasterize pop data (100x100 m) to aWhere grid (9x9 km)
        >>> vt_pop_rasterized = awg.rasterize(vt_grid, vt_pop_path)
        >>> # Display single entry in resulting geodataframe
        >>> vt_pop_rasterized.loc[1]
        geometry    POLYGON ((-73.4778398510704 43.56701241643375,...
        centroid         POINT (-73.43783985107041 43.60701241643375)
        count                                                    2759
        sum                                                   12.7015
        Name: 1, dtype: object
    """
    # Open data for rasterizing
    with rio.open(raster_path) as src:

        # Extract array and metadata
        src_arr = src.read(1, masked=True)
        src_meta = src.profile

    # Create copy of input geodataframe
    awhere_grid_copy = awhere_grid.copy()

    # Extract zonal stats (list of dictionaries)
    geojson_list = rs.zonal_stats(
        awhere_grid_copy,
        src_arr,
        nodata=src_meta.get("nodata"),
        affine=src_meta.get("transform"),
        geojson_out=True,
        copy_properties=True,
        stats=zonal_stats,
    )

    # Convert list to geodataframe
    awhere_grid_rasterized = gpd.GeoDataFrame.from_features(geojson_list)

    # Return rasterized geodataframe
    return awhere_grid_rasterized


def plot_grid(
    awhere_grid, study_area, plot_title="aWhere Grid", data_source=None
):
    """Plots a administrative boundary with the
    mathching awhere grid.

    Parameters
    ----------
    awhere_grid : geopandas geodataframe
        Geodataframe containing a grid of aWhere-sized
        cells.

    study_area : geopandas geodataframe
        Geodataframe containing the study area boundary
        polygon.

    plot_title : str
        Title for the plot. Default value is 'aWhere Grid'.

    Returns
    -------
    tuple

        fig: matplotlib.figure.Figure object
            The figure object associated with the histogram.

        ax: matplotlib.axes._subplots.AxesSubplot objects
            The axes object associated with the histogram.

    Example
    -------
        >>> # Import packages
        >>> import os
        >>> import awherepy.grids as awg
        >>> # Define path to shapefile boundary
        >>> vt_bound_path = os.path.join(
        ...     working_directory, 'shapefiles', vermont_state_boundary.shp')
        >>> # Create aWhere grid
        >>> vt_grid, vt_bound_4326 = create_grid(
        ...     vt_bound_path, buffer_distance=0.12)
        >>> # Plot grid and boundary
        >>> fig, ax = plot_grid(vt_grid, vt_bound_4326)
    """
    # Project to WGS 84 Lat/Lon, EPSG 4326 if no CRS match
    if not study_area.crs == "epsg:4326":
        study_area_4326 = study_area.to_crs("epsg:4326")

    else:
        study_area_4326 = study_area

    # Use dark background
    with plt.style.context("dark_background"):

        # Define figure and axes
        fig, ax = plt.subplots(figsize=(20, 10))

        # Plot awhere grid and study area on same axes
        awhere_grid.plot(
            ax=ax, facecolor="none", edgecolor="#984ea3", linewidth=1.5
        )

        study_area_4326.plot(
            ax=ax, facecolor="none", edgecolor="#4daf4a", linewidth=2
        )

        # Configures axes
        ax.set_xlabel("Longitude (degrees)", fontsize=16)
        ax.set_ylabel("Latitude (degrees)", fontsize=16)
        ax.set_title(plot_title, fontsize=20)
        ax.tick_params(axis="both", which="major", labelsize=16)

        # Add caption if data source included
        if data_source:
            fig.text(
                0.5,
                0.035,
                f"Data Source: {data_source}",
                ha="center",
                fontsize=12,
            )

    # Return figure and axes
    return fig, ax


def export_grid(awhere_grid, output_path):
    """Exports an aWhere grid to the specified file format.

    Functionality currently supports writing to
    comma-separated values (.CSV), shapefile (.SHP),
    geojson (.GEOJSON), geopackage (.GPKG).

    Parameters
    ----------
    awhere_grid : geopandas geodataframe
        Geodataframe containing a grid of aWhere-sized
        cells.

    output_path : str
        Path to write the output file (including file
        name and extension).

    Returns
    -------
    output_message : str
        Message indicating successful export (and path)
        or error.

    Example
    -------
        >>> # Import packages
        >>> import os
        >>> import awherepy.grids as awg
        >>> # Define path to shapefile boundary
        >>> vt_bound_path = os.path.join(
        ...     working_directory, 'shapefiles',
        ...     vermont_state_boundary.shp')
        >>> # Create aWhere grid
        >>> vt_grid, vt_bound_4326 = awg.create_grid(
        ...     vt_bound_path, buffer_distance=0.12)
        >>> # Define export path
        >>> # export_path = os.path.join(
        ...     working_directory, '03-processed-data',
        ...     'vt_grid.csv')
        >>> # Export grid to CSV
        >>> csv_export = awg.export_grid(vt_grid, export_path)
        vt_grid.csv
    """
    # Determine output folder
    output_folder = os.path.dirname(output_path)

    # Raise error if output folder does not exist
    if not os.path.exists(output_folder):

        raise OSError("Directory does not exist.")

    # Extract file extension
    file_extension = output_path.split(".")[-1]

    # Define drivers for export
    drivers = {"shp": "ESRI Shapefile", "geojson": "GeoJSON", "gpkg": "GPKG"}

    # Check file extension; determine export method
    if file_extension == "csv":

        # Write to CSV
        awhere_grid.to_csv(path_or_buf=output_path, index=False)
        output_message = output_path

    elif file_extension.lower() in drivers.keys():

        # Write to shp, geojson, or gpkg
        awhere_grid.to_file(output_path, driver=drivers.get(file_extension))
        output_message = output_path

    else:
        # Raise error for unsupported file type
        raise ValueError(
            f"Invalid output format: .{file_extension}"
            "\nFormats supported: .csv, .shp, .geojson, and .gpkg"
        )

    # Return output message
    return output_message



"""
awherepy.crops
==============
A module to access and retrieve aWhere crop data.
"""

import requests as rq
import pandas as pd
from pandas.io.json import json_normalize
import awherepy as aw

# Define valid crops
CROPS_LIST = [
    "barley-generic",
    "canola-b-napus",
    "canola-b-rapa",
    "canola-generic",
    "corn-2300-gdd",
    "corn-2500-gdd",
    "corn-2700-gdd",
    "corn-2900-gdd",
    "cotton-2200-gdd",
    "cotton-2600-gdd",
    "green-gram-generic",
    "oat-generic",
    "potato-generic",
    "sorghum-long-season",
    "sorghum-short-season",
    "soya-generic",
    "sugarbeet-generic",
    "sunflower-early-dwarf",
    "sunflower-generic",
    "wheat-generic",
    "wheat-hardred",
]

# Define variables for data cleaning
CROPS_DROP_COLS = [
    "_links.self.href",
    "_links.curies",
    "_links.awhere:plantings.href",
]

CROPS_RENAME_MAP = {
    "id": "crop_id",
    "name": "crop_name",
    "type": "crop_type",
    "variety": "crop_variety",
    "isDefaultForCrop": "default_crop",
}


def get_crops(
    key, secret, crop_id=None,
):
    """Retrieve a list of available crops or single crop if specified.

    API reference:
        https://docs.awhere.com/knowledge-base-docs/crop-list-detail/

    Parameters
    ----------
    key : str
        API key for a valid aWhere API application.

    secret : str
        API secret for a valid aWhere API application.

    crop_id : str, optional
        Valid aWhere crop id.

    Returns
    -------
    crop_df : pandas dataframe
        Dataframe containing a single crop or all available crops.

    Example
    -------
        >>> # Imports
        >>> import os
        >>> import awherepy as aw
        >>> import awherepy.crops as awc
        >>> # Get aWhere API key and secret
        >>> awhere_api_key = os.environ.get('AWHERE_API_KEY')
        >>> awhere_api_secret = os.environ.get('AWHERE_API_SECRET')
        >>> # Get all available aWhere crops
        >>> all_crops = awc.get_crops(awhere_api_key, awhere_api_secret)
        >>> # Check number of entries
        >>> len(all_crops)
        21
        >>> # Get a specific crop
        >>> corn_2500_gdd = awc.get_crops(
        >>>     awhere_api_key, awhere_api_secret, crop_id='corn-2500-gdd'
        >>> )
        >>> Check number of entries
        >>> len(corn_2500_gdd)
        1
    """
    # Define global variables
    global CROPS_LIST, CROPS_DROP_COLS, CROPS_RENAME_MAP

    # Define crops api url
    api_url = "https://api.awhere.com/v2/agronomics/crops"

    # Check if credentials are valid
    if aw.valid_credentials(key, secret):

        # Get OAuth token
        auth_token = aw.get_oauth_token(key, secret)

        # Set up the HTTP request headers
        auth_headers = {"Authorization": f"Bearer {auth_token}"}

        # Check if single crop
        if crop_id:

            # Raise error if crop id is invalid
            if crop_id not in CROPS_LIST:
                raise KeyError(
                    (
                        f"Invalid crop id."
                        f"Valid crop ids: {', '.join(CROPS_LIST)}"
                    )
                )

            # Get API response
            response = rq.get(f"{api_url}/{crop_id}", headers=auth_headers)

            # Convert API response to JSON format
            response_json = response.json()

            # Convert to dataframe
            crop_df = json_normalize(response_json)

        # All crops
        else:

            # Initialize variables
            offset = 0
            limit = 10
            pages = 3

            # Initialize list to store dataframe for each page of results
            response_df_list = []

            # Loop through all pages
            while offset < limit * pages:

                # Get API response
                response = rq.get(
                    f"{api_url}?limit={limit}&offset={offset}",
                    headers=auth_headers,
                )

                # Convert API response to JSON format
                response_json = response.json()

                # Convert json to dataframe
                response_df_loop = json_normalize(response_json.get("crops"))

                # Append to dataframe list
                response_df_list.append(response_df_loop)

                # Increase offset by 1 page of results
                offset += limit

            # Merge all page dataframes into a single dataframe
            crop_df = pd.concat(response_df_list, axis=0)

        # Drop unnecessary columns
        crop_df.drop(
            columns=CROPS_DROP_COLS, inplace=True,
        )

        # Reset index
        crop_df.reset_index(drop=True, inplace=True)

        # Rename columns
        crop_df.rename(columns=CROPS_RENAME_MAP, inplace=True)

    # Invalid credentials
    else:
        # Raise error
        raise ValueError("Invalid aWhere API credentials.")

    # Return crops
    return crop_df



"""
awherepy.agronomics
===================
A module to access and retrieve aWhere agronomics data.
"""

import requests as rq
from datetime import date
import pandas as pd
from pandas.io.json import json_normalize
import geopandas as gpd
import awherepy as aw

# Define variables for data cleaning - norms
# Single day
NORMS_SINGLE_DAY_COORD_COLS = ["location.longitude", "location.latitude"]

NORMS_SINGLE_DAY_DROP_COLS = [
    "location.fieldId",
    "pet.units",
    "_links.self.href",
    "_links.curies",
    "_links.awhere:field.href",
]

NORMS_SINGLE_DAY_RENAME_MAP = {
    "gdd.average": "gdd_daily_average_total_cels",
    "gdd.stdDev": "gdd_daily_average_total_std_dev_cels",
    "pet.average": "pet_daily_average_total_mm",
    "pet.stdDev": "pet_daily_average_total_std_dev_mm",
    "ppet.average": "ppet_daily_average_total",
    "ppet.stdDev": "ppet_daily_average_total_std_dev",
}

# Multi-day
NORMS_MULTIPLE_DAY_COORD_COLS = ["longitude", "latitude"]

# Multi-day, total accumulation
NORMS_MULTIPLE_DAY_TOTAL_DROP_COLS = ["precipitation.units", "pet.units"]

NORMS_MULTIPLE_DAY_TOTAL_RENAME_MAP = {
    "gdd.average": "gdd_range_average_total_cels",
    "gdd.stdDev": "gdd_range_average_total_std_dev_cels",
    "precipitation.average": "precip_range_average_total_mm",
    "precipitation.stdDev": "precip_range_average_total_std_dev_mm",
    "pet.average": "pet_range_average_total_mm",
    "pet.stdDev": "pet_range_average_total_std_dev",
    "ppet.average": "ppet_range_daily_average",
    "ppet.stdDev": "ppet_range_daily_average_std_dev",
}

# Multi-day, daily accumulation
NORMS_MULTIPLE_DAY_DAILY_DROP_COLS = [
    "pet.units",
    "accumulatedPrecipitation.units",
    "accumulatedPet.units",
    "_links.self.href",
    "_links.curies",
    "_links.awhere:field.href",
]

NORMS_MULTIPLE_DAY_DAILY_RENAME_MAP = {
    "gdd.average": "gdd_daily_average_cels",
    "gdd.stdDev": "gdd_daily_average_std_dev_cels",
    "pet.average": "pet_daily_average_mm",
    "pet.stdDev": "pet_daily_average_std_dev_mm",
    "ppet.average": "ppet_daliy_average",
    "ppet.stdDev": "ppet_daily_average_std_dev",
    "accumulatedGdd.average": "gdd_rolling_total_average",
    "accumulatedGdd.stdDev": "gdd_rolling_total_average_std_dev",
    "accumulatedPrecipitation.average": "precip_rolling_total_average_mm",
    "accumulatedPrecipitation.stdDev": (
        "precip_rolling_total_average_" "std_dev_mm"
    ),
    "accumulatedPet.average": "pet_rolling_total_average_mm",
    "accumulatedPet.stdDev": "pet_rolling_total_average_std_dev_mm",
    "accumulatedPpet.average": "ppet_rolling_total_average",
    "accumulatedPpet.stdDev": "ppet_rolling_total_average_std_dev",
}

# Define variables for data cleaning - values
# Single day
VALUES_SINGLE_DAY_COORD_COLS = ["location.longitude", "location.latitude"]

VALUES_SINGLE_DAY_DROP_COLS = [
    "location.fieldId",
    "pet.units",
    "_links.self.href",
    "_links.curies",
    "_links.awhere:field.href",
]

VALUES_SINGLE_DAY_RENAME_MAP = {
    "gdd": "gdd_daily_total_cels",
    "ppet": "ppet_daily_total",
    "pet.amount": "pet_daily_total_mm",
}


# Multi-day
VALUES_MULTIPLE_DAY_COORD_COLS = ["longitude", "latitude"]


# Multi-day, total accumulation
VALUES_MULTIPLE_DAY_TOTAL_DROP_COLS = ["precipitation.units", "pet.units"]

VALUES_MULTIPLE_DAY_TOTAL_RENAME_MAP = {
    "gdd": "gdd_range_total_cels",
    "ppet": "ppet_range_total",  # This is accumulation of all daily PPET
    "precipitation.amount": "precip_range_total_mm",
    "pet.amount": "pet_range_total_mm",
}

# Multi-day, daily accumulation
VALUES_MULTIPLE_DAY_DAILY_DROP_COLS = [
    "pet.units",
    "accumulatedPrecipitation.units",
    "accumulatedPet.units",
    "_links.self.href",
    "_links.curies",
    "_links.awhere:field.href",
]

VALUES_MULTIPLE_DAY_DAILY_RENAME_MAP = {
    "gdd": "gdd_daily_total_cels",
    "ppet": "ppet_daily_total",
    "accumulatedGdd": "gdd_rolling_total_cels",
    "accumulatedPpet": "ppet_rolling_total",
    "pet.amount": "pet_daily_total_mm",
    "accumulatedPrecipitation.amount": "precip_rolling_total_mm",
    "accumulatedPet.amount": "pet_rolling_total_mm",
}


def _call_agronomic_norms(
    key,
    secret,
    input_type="location",
    location=(-105.648222, 40.313250),
    field_id=None,
    start_date="01-01",
    end_date=None,
    limit=10,
    offset=0,
):
    """Helper function that calls the aWhere API for agronomic norms.
    """
    # Check input data type
    # Location-based
    if input_type == "location":

        # Raise error if location is not defined
        if location is None:
            raise ValueError(
                ("Must specify a location, with longitude " "and latitude.")
            )

        # Set URL to location
        api_url = (
            "https://api.awhere.com/v2/agronomics/locations/"
            f"{location[1]},{location[0]}"
        )

    # Field-based
    elif input_type == "field":

        # Raise error if field name is not defined
        if field_id is None:
            raise ValueError("Must specify a field name.")

        # Set URL to fields
        api_url = f"https://api.awhere.com/v2/agronomics/fields/{field_id}"

    # Invalid input
    else:
        raise ValueError("Invalid input type. Must be 'location' or 'field'.")

    # Get OAuth token
    auth_token = aw.get_oauth_token(key, secret)

    # Setup the HTTP request headers
    auth_headers = {"Authorization": f"Bearer {auth_token}"}

    # Define URL variants
    url_single_day = f"{api_url}/agronomicnorms/{start_date}"
    url_multiple_days = (
        f"{api_url}/agronomicnorms/{start_date},{end_date}"
        f"?limit={limit}&offset={offset}"
    )

    # Get single day norms or date range
    response = (
        rq.get(url_multiple_days, headers=auth_headers)
        if end_date
        else rq.get(url_single_day, headers=auth_headers)
    )

    # Convert response to json format
    agronomic_norms = response.json()

    # Return norms
    return agronomic_norms


def _extract_agronomic_norms(agronomic_norms):
    """Helper function that extracts agronomic norms into a dataframe.
    """
    # Extract lat/lon
    latitude = agronomic_norms.get("location").get("latitude")
    longitude = agronomic_norms.get("location").get("longitude")

    # Check if more than one day
    if agronomic_norms.get("dailyNorms"):

        # DAILY ACCUMULATION NORMS
        # Get daily accumulation norms
        daily_norms = json_normalize(agronomic_norms.get("dailyNorms"))

        # Add lat/lon and set date as index
        daily_norms["latitude"] = latitude
        daily_norms["longitude"] = longitude
        daily_norms.set_index(["day"], inplace=True)

        # TOTAL ACCUMULATION NORMS
        # Get average accumulations through all days
        total_norms = json_normalize(
            agronomic_norms.get("averageAccumulations")
        )

        # Get list of dates, add start/end dates, set date range as index
        dates = [
            entry.get("day") for entry in agronomic_norms.get("dailyNorms")
        ]

        # Populate date columns
        total_norms["date_range"] = f"{dates[0]}/{dates[-1]}"
        total_norms["start_day"] = dates[0]
        total_norms["end_day"] = dates[-1]
        total_norms.set_index(["date_range"], inplace=True)

        # Add lat/lon
        total_norms["latitude"] = latitude
        total_norms["longitude"] = longitude

        # Put dataframes in tuple (total norms, daily norms)
        agronomics_df = (total_norms, daily_norms)

    # Single day
    else:
        agronomics_df = json_normalize(agronomic_norms)
        agronomics_df.set_index(["day"], inplace=True)

    return agronomics_df


def _clean_agronomic_norms(df):
    """Helper function that cleans and georeferences agronomics norms.
    """
    # Define global variables
    global NORMS_SINGLE_DAY_COORD_COLS
    global NORMS_SINGLE_DAY_DROP_COLS
    global NORMS_SINGLE_DAY_RENAME_MAP
    global NORMS_MULTIPLE_DAY_COORD_COLS
    global NORMS_MULTIPLE_DAY_TOTAL_DROP_COLS
    global NORMS_MULTIPLE_DAY_TOTAL_RENAME_MAP
    global NORMS_MULTIPLE_DAY_DAILY_DROP_COLS
    global NORMS_MULTIPLE_DAY_DAILY_RENAME_MAP

    # Define CRS (EPSG 4326)
    crs = "epsg:4326"

    # Check type of input data
    # Pandas dataframe (single-day input)
    if isinstance(df, pd.core.frame.DataFrame):

        # Create copy of input dataframe; prevents altering the original
        df_copy = df.copy()

        # Convert to geodataframe
        gdf = gpd.GeoDataFrame(
            df_copy,
            crs=crs,
            geometry=gpd.points_from_xy(
                df_copy[NORMS_SINGLE_DAY_COORD_COLS[0]],
                df_copy[NORMS_SINGLE_DAY_COORD_COLS[1]],
            ),
        )

        # Add lat/lon columns to drop columns list
        NORMS_SINGLE_DAY_DROP_COLS += NORMS_SINGLE_DAY_COORD_COLS

        # Drop columns
        gdf.drop(
            columns=NORMS_SINGLE_DAY_DROP_COLS,
            axis=1,
            inplace=True,
            errors="ignore",
        )

        # Rename columns
        gdf.rename(
            columns=NORMS_SINGLE_DAY_RENAME_MAP, inplace=True, errors="ignore"
        )

    # Tuple of dataframes (multi-day input)
    elif isinstance(df, tuple):

        # Total accumulation
        # Create copy of input dataframe; prevents altering the original
        total_df_copy = df[0].copy()

        # Convert to geodataframe
        total_gdf = gpd.GeoDataFrame(
            total_df_copy,
            crs=crs,
            geometry=gpd.points_from_xy(
                total_df_copy[NORMS_MULTIPLE_DAY_COORD_COLS[0]],
                total_df_copy[NORMS_MULTIPLE_DAY_COORD_COLS[1]],
            ),
        )

        # Add lat/lon columns to drop columns list
        NORMS_MULTIPLE_DAY_TOTAL_DROP_COLS += NORMS_MULTIPLE_DAY_COORD_COLS

        # Drop columns
        total_gdf.drop(
            columns=NORMS_MULTIPLE_DAY_TOTAL_DROP_COLS,
            axis=1,
            inplace=True,
            errors="ignore",
        )

        # Rename columns
        total_gdf.rename(
            columns=NORMS_MULTIPLE_DAY_TOTAL_RENAME_MAP,
            inplace=True,
            errors="ignore",
        )

        # Daily accumulation
        # Create copy of input dataframe; prevents altering the original
        daily_df_copy = df[1].copy()

        # Convert to geodataframe
        daily_gdf = gpd.GeoDataFrame(
            daily_df_copy,
            crs=crs,
            geometry=gpd.points_from_xy(
                daily_df_copy[NORMS_MULTIPLE_DAY_COORD_COLS[0]],
                daily_df_copy[NORMS_MULTIPLE_DAY_COORD_COLS[1]],
            ),
        )

        # Add lat/lon columns to drop columns list
        NORMS_MULTIPLE_DAY_DAILY_DROP_COLS += NORMS_MULTIPLE_DAY_COORD_COLS

        # Drop columns
        daily_gdf.drop(
            columns=NORMS_MULTIPLE_DAY_DAILY_DROP_COLS,
            axis=1,
            inplace=True,
            errors="ignore",
        )

        # Rename columns
        daily_gdf.rename(
            columns=NORMS_MULTIPLE_DAY_DAILY_RENAME_MAP,
            inplace=True,
            errors="ignore",
        )

        # Put dataframes in tuple (total norms, daily norms)
        gdf = (total_gdf, daily_gdf)

    # Invalid input type
    else:
        # Raise error
        raise TypeError("Invalid input type.")

    # Return cleaned up geodataframe
    return gdf


def get_agronomic_norms(key, secret, kwargs=None):
    """Gets historical agronomic norms data from the aWhere API.

    API reference: https://docs.awhere.com/knowledge-base-docs/agronomics/

    Parameters
    ----------
    key : str
        API key for a valid aWhere API application.

    secret : str
        API secret for a valid aWhere API application.

    kwargs : dict, optional
        Keyword arguments for different query parameters.
        Running the function without kwargs will use the
        default values. Arguments include:

            input_type: str
                Type of data, agronomics API by geolocation or API by field.
                Valid options include 'location' and 'field'. Default
                value is 'location'.

            location: tuple of float
                Tuple containing the location (longitude, latitude) that the
                API gets data from. For use with the 'location' input_type.
                Default value is (-105.648222, 40.313250), for Bear Lake, CO.

            field_id: str
                Field ID for a valid aWhere field associated with the input API
                key/secret. For use with the 'field' input_type. Default value
                is None.

            start_date: str
                Start date for the agronomic norms data to be retrieved.
                Formatted 'MM-DD'. All days of the year are valid, from
                '01-01' to '12-31'. Default value is '01-01'.

            end_date: str
                End date for the agronomic norms data to be retrieved.
                Formatted 'MM-DD'. All days of the year are valid, from
                '01-01' to '12-31'. Default value is None.

            limit: int
                Number of results in the API response per page. Used with
                offset kwarg to sort through pages of results (cases where the
                number of results exceeds the limit per page). Applicable when
                the number of results exceeds 1. Maximum value is 10. Default
                value is 10.

            offset: int
                Number of results in the API response to skip. Used with limit
                kwarg to sort through pages of results (cases where the
                number of results exceeds the limit per page). Applicable when
                the number of results exceeds 1. Default value is 0 (start
                with first result).

    Returns
    -------
    norms_gdf : geopandas geodataframe or tuple of geodataframes
        Geotdataframe(s) containing the historical agronomic norms
        data for the specified location or aWhere field, for the
        specified date(s). Single-day inputs return a single
        geodataframe. Multi-day inputs return a tuple of geodataframes
        (total norms, daily norms).

    Example
    -------
        >>> # Imports
        >>> import os
        >>> import awherepy as aw
        >>> import awherepy.agronomics as awa
        >>> # Get aWhere API key and secret
        >>> awhere_api_key = os.environ.get('AWHERE_API_KEY')
        >>> awhere_api_secret = os.environ.get('AWHERE_API_SECRET')
        >>> # Get historical norms with default kwargs (Bear Lake, CO)
        >>> bear_lake_norms = awa.get_agronomic_norms(
        ...     key=awhere_api_key, secret=awhere_api_secret)
        >>> # Check number of entries in geodataframe
        >>> len(bear_lake_norms)
        1
        >>> # Define non-default kwargs (Manchester Center, Vermont)
        >>> # Define kwargs for Manchester, Vermont
        >>> vt_kwargs = {
        ...     'location': (-73.0723269, 43.1636875),
        ...     'start_date': '05-10',
        ...     'end_date': '05-19'
        ... }
        >>> # Get historical norms for Manchester, Vermont
        >>> vt_total_norms, vt_daily_norms = awa.get_agronomic_norms(
        ...     key=awhere_api_key,
        ...     secret=awhere_api_secret,
        ...     kwargs=vt_kwargs
        ... )
        >>> # Check number entries in geodataframe
        >>> len(vt_total_norms)
        10
    """
    # Check if credentials are valid
    if aw.valid_credentials(key, secret):

        # Call data
        norms_json = (
            _call_agronomic_norms(key, secret, **kwargs)
            if kwargs
            else _call_agronomic_norms(key, secret)
        )

        # Extract data
        norms_df = _extract_agronomic_norms(norms_json)

        # Clean data
        norms_gdf = _clean_agronomic_norms(norms_df)

    else:
        # Raise error
        raise ValueError("Invalid aWhere API credentials.")

    # Return cleaned data
    return norms_gdf


def _call_agronomic_values(
    key,
    secret,
    input_type="location",
    location=(-105.648222, 40.313250),
    field_id=None,
    start_date=date.today().strftime("%m-%d"),
    end_date=None,
    limit=10,
    offset=0,
):
    """Helper function that calls the aWhere API for agronomic values.
    """
    # Check input data type
    # Location-based
    if input_type == "location":

        # Raise error if location is not defined
        if location is None:
            raise ValueError(
                ("Must specify a location, with longitude " "and latitude.")
            )

        # Set URL to location
        api_url = (
            "https://api.awhere.com/v2/agronomics/locations/"
            f"{location[1]},{location[0]}"
        )

    # Field-based
    elif input_type == "field":

        # Raise error if field name is not defined
        if field_id is None:
            raise ValueError("Must specify a field name.")

        # Set URL to fields
        api_url = f"https://api.awhere.com/v2/agronomics/fields/{field_id}"

    # Invalid input
    else:
        raise ValueError("Invalid input type. Must be 'location' or 'field'.")

    # Get OAuth token
    auth_token = aw.get_oauth_token(key, secret)

    # Setup the HTTP request headers
    auth_headers = {"Authorization": f"Bearer {auth_token}"}

    # Define URL variants
    url_single_day = f"{api_url}/agronomicvalues/{start_date}"
    url_multiple_days = (
        f"{api_url}/agronomicvalues/{start_date},{end_date}"
        f"?limit={limit}&offset={offset}"
    )

    # Get single day norms or date range
    response = (
        rq.get(url_multiple_days, headers=auth_headers)
        if end_date
        else rq.get(url_single_day, headers=auth_headers)
    )

    # Convert response to json format
    agronomic_values = response.json()

    # Return norms
    return agronomic_values


def _extract_agronomic_values(agronomic_values):
    """Helper function that extracts agronomic values into a dataframe.
    """
    # Extract lat/lon
    latitude = agronomic_values.get("location").get("latitude")
    longitude = agronomic_values.get("location").get("longitude")

    # Check if more than one day
    if agronomic_values.get("dailyValues"):

        # DAILY ACCUMULATIONS
        # Get daily forecasted accumulations
        daily_accumulation = json_normalize(
            agronomic_values.get("dailyValues")
        )

        # Add lat/lon and set date as index
        daily_accumulation["latitude"] = latitude
        daily_accumulation["longitude"] = longitude
        daily_accumulation.set_index(["date"], inplace=True)

        # TOTAL ACCUMULATION
        # Get total forecasted accumulations through all days
        total_accumulation = json_normalize(
            agronomic_values.get("accumulations")
        )

        # Get list of dates, add start/end dates, set date range as index
        dates = [
            entry.get("date") for entry in agronomic_values.get("dailyValues")
        ]

        # Populate date columns
        total_accumulation["date_range"] = f"{dates[0]}/{dates[-1]}"
        total_accumulation["start_day"] = dates[0]
        total_accumulation["end_day"] = dates[-1]
        total_accumulation.set_index(["date_range"], inplace=True)

        # Add lat/lon
        total_accumulation["latitude"] = latitude
        total_accumulation["longitude"] = longitude

        # Put dataframes in tuple (total accumulation, daily accumulation)
        agronomics_df = (total_accumulation, daily_accumulation)

    # Single day
    else:
        agronomics_df = json_normalize(agronomic_values)
        agronomics_df.set_index(["date"], inplace=True)

    return agronomics_df


def _clean_agronomic_values(df):
    """Helper function that cleans and georeferences agronomics values.
    """
    # Define global variables
    global VALUES_SINGLE_DAY_COORD_COLS
    global VALUES_SINGLE_DAY_DROP_COLS
    global VALUES_SINGLE_DAY_RENAME_MAP
    global VALUES_MULTIPLE_DAY_COORD_COLS
    global VALUES_MULTIPLE_DAY_TOTAL_DROP_COLS
    global VALUES_MULTIPLE_DAY_TOTAL_RENAME_MAP
    global VALUES_MULTIPLE_DAY_DAILY_DROP_COLS
    global VALUES_MULTIPLE_DAY_DAILY_RENAME_MAP

    # Define CRS (EPSG 4326)
    crs = "epsg:4326"

    # Check type of input data
    # Pandas dataframe (single-day input)
    if isinstance(df, pd.core.frame.DataFrame):

        # Create copy of input dataframe; prevents altering the original
        df_copy = df.copy()

        # Convert to geodataframe
        gdf = gpd.GeoDataFrame(
            df_copy,
            crs=crs,
            geometry=gpd.points_from_xy(
                df_copy[VALUES_SINGLE_DAY_COORD_COLS[0]],
                df_copy[VALUES_SINGLE_DAY_COORD_COLS[1]],
            ),
        )

        # Add lat/lon columns to drop columns list
        VALUES_SINGLE_DAY_DROP_COLS += VALUES_SINGLE_DAY_COORD_COLS

        # Drop columns
        gdf.drop(
            columns=VALUES_SINGLE_DAY_DROP_COLS,
            axis=1,
            inplace=True,
            errors="ignore",
        )

        # Rename columns
        gdf.rename(
            columns=VALUES_SINGLE_DAY_RENAME_MAP, inplace=True, errors="ignore"
        )

    # Tuple of dataframes (multi-day input)
    elif isinstance(df, tuple):

        # Total accumulation
        # Create copy of input dataframe; prevents altering the original
        total_df_copy = df[0].copy()

        # Convert to geodataframe
        total_gdf = gpd.GeoDataFrame(
            total_df_copy,
            crs=crs,
            geometry=gpd.points_from_xy(
                total_df_copy[VALUES_MULTIPLE_DAY_COORD_COLS[0]],
                total_df_copy[VALUES_MULTIPLE_DAY_COORD_COLS[1]],
            ),
        )

        # Add lat/lon columns to drop columns list
        VALUES_MULTIPLE_DAY_TOTAL_DROP_COLS += VALUES_MULTIPLE_DAY_COORD_COLS

        # Drop columns
        total_gdf.drop(
            columns=VALUES_MULTIPLE_DAY_TOTAL_DROP_COLS,
            axis=1,
            inplace=True,
            errors="ignore",
        )

        # Rename columns
        total_gdf.rename(
            columns=VALUES_MULTIPLE_DAY_TOTAL_RENAME_MAP,
            inplace=True,
            errors="ignore",
        )

        # Daily accumulation
        # Create copy of input dataframe; prevents altering the original
        daily_df_copy = df[1].copy()

        # Convert to geodataframe
        daily_gdf = gpd.GeoDataFrame(
            daily_df_copy,
            crs=crs,
            geometry=gpd.points_from_xy(
                daily_df_copy[VALUES_MULTIPLE_DAY_COORD_COLS[0]],
                daily_df_copy[VALUES_MULTIPLE_DAY_COORD_COLS[1]],
            ),
        )

        # Add lat/lon columns to drop columns list
        VALUES_MULTIPLE_DAY_DAILY_DROP_COLS += VALUES_MULTIPLE_DAY_COORD_COLS

        # Drop columns
        daily_gdf.drop(
            columns=VALUES_MULTIPLE_DAY_DAILY_DROP_COLS,
            axis=1,
            inplace=True,
            errors="ignore",
        )

        # Rename columns
        daily_gdf.rename(
            columns=VALUES_MULTIPLE_DAY_DAILY_RENAME_MAP,
            inplace=True,
            errors="ignore",
        )

        # Put dataframes in tuple (total norms, daily norms)
        gdf = (total_gdf, daily_gdf)

    # Invalid input type
    else:
        # Raise error
        raise TypeError("Invalid input type.")

    # Return cleaned up geodataframe
    return gdf


def get_agronomic_values(key, secret, kwargs=None):
    """Gets historical agronomic values (forecast and observed)
    data from the aWhere API.

    API reference: https://docs.awhere.com/knowledge-base-docs/agronomics/

    Parameters
    ----------
    key : str
        API key for a valid aWhere API application.

    secret : str
        API secret for a valid aWhere API application.

    kwargs : dict, optional
        Keyword arguments for different query parameters.
        Running the function without kwargs will use the
        default values. Arguments include:

            input_type: str
                Type of data, agronomics API by geolocation or API by field.
                Valid options include 'location' and 'field'. Default
                value is 'location'.

            location: tuple of float
                Tuple containing the location (longitude, latitude) that the
                API gets data from. For use with the 'location' input_type.
                Default value is (-105.648222, 40.313250), for Bear Lake, CO.

            field_id: str
                Field ID for a valid aWhere field associated with the input API
                key/secret. For use with the 'field' input_type. Default value
                is None.

            start_date: str
                Start date for the agronomic values data to be retrieved.
                Formatted 'YYYY-MM-DD'. All days of the year are valid,
                from 'YYYY-01-01' to 'YYYY-12-31'. Default value is the
                current day.

            end_date: str
                End date for the agronomic values data to be retrieved.
                Formatted 'MM-DD'. All days of the year are valid, from
                'YYYY-01-01' to 'YYYY-12-31'. Default value is None.

            limit: int
                Number of results in the API response per page. Used with
                offset kwarg to sort through pages of results (cases where the
                number of results exceeds the limit per page). Applicable when
                the number of results exceeds 1. Maximum value is 10. Default
                value is 10.

            offset: int
                Number of results in the API response to skip. Used with limit
                kwarg to sort through pages of results (cases where the
                number of results exceeds the limit per page). Applicable when
                the number of results exceeds 1. Default value is 0 (start
                with first result).

    Returns
    -------
    values_gdf : geopandas geodataframe or tuple of geodataframes
        Geotdataframe(s) containing the agronomics data (forecast
        or observed) data for the specified location or aWhere field,
        for the specified date(s). Single-day inputs return a single
        geodataframe. Multi-day inputs return a tuple of geodataframes
        (total values, daily values).

    Example
    -------
        >>> # Imports
        >>> import os
        >>> import awherepy as aw
        >>> import awherepy.agronomics as awa
        >>> # Get aWhere API key and secret
        >>> awhere_api_key = os.environ.get('AWHERE_API_KEY')
        >>> awhere_api_secret = os.environ.get('AWHERE_API_SECRET')
        >>> # Get historical norms with default kwargs (Bear Lake, CO)
        >>> bear_lake_values = awa.get_agronomic_values(
        ...     key=awhere_api_key, secret=awhere_api_secret)
        >>> # Check number of entries in geodataframe
        >>> len(bear_lake_norms)
        1
        >>> # Define non-default kwargs (Manchester Center, Vermont)
        >>> # Define kwargs for Manchester, Vermont
        >>> vt_kwargs = {
        ...     'location': (-73.0723269, 43.1636875),
        ...     'start_date': '2020-05-10',
        ...     'end_date': '2020-05-19'
        ... }
        >>> # Get historical norms for Manchester, Vermont
        >>> vt_total_values, vt_daily_values = awa.get_agronomic_values(
        ...     key=awhere_api_key,
        ...     secret=awhere_api_secret,
        ...     kwargs=vt_kwargs
        ... )
        >>> # Check number entries in geodataframe
        >>> len(vt_total_norms)
        10
    """
    # Check if credentials are valid
    if aw.valid_credentials(key, secret):

        # Call data
        values_json = (
            _call_agronomic_values(key, secret, **kwargs)
            if kwargs
            else _call_agronomic_values(key, secret)
        )

        # Extract data
        values_df = _extract_agronomic_values(values_json)

        # Clean data
        values_gdf = _clean_agronomic_values(values_df)

    else:
        # Raise error
        raise ValueError("Invalid aWhere API credentials.")

    # Return cleaned data
    return values_gdf



from os import path
from setuptools import setup

DISTNAME = "awherepy"
DESCRIPTION = (
    "Python package built to work with weather and agriculture "
    "data in the aWhere API."
)
MAINTAINER = "Cale Kochenour"
MAINTAINER_EMAIL = "cale.kochenour@alumni.psu.edu"


this_directory = path.abspath(path.dirname(__file__))
with open(path.join(this_directory, "README.md"), encoding="utf-8") as f:
    LONG_DESCRIPTION = f.read()

if __name__ == "__main__":
    setup(
        name=DISTNAME,
        maintainer=MAINTAINER,
        maintainer_email=MAINTAINER_EMAIL,
        description=DESCRIPTION,
        long_description=LONG_DESCRIPTION,
        long_description_content_type="text/markdown",
        version="0.1.0",
        packages=["awherepy"],
        install_requires=[
            "matplotlib",
            "numpy",
            "seaborn",
            "geopandas",
            "earthpy",
            "shapely",
            "rasterstats",
            "descartes",
            "rtree",
        ],
        zip_safe=False,
        classifiers=[
            "Intended Audience :: Developers",
            "License :: OSI Approved :: BSD License",
            "Programming Language :: Python",
            "Topic :: Software Development",
            "Operating System :: Microsoft :: Windows",
            "Operating System :: POSIX",
            "Operating System :: Unix",
            "Operating System :: MacOS",
        ],
        package_data={
            DISTNAME: [
                "tests/test-data/*.cpg",
                "tests/test-data/*.dbf",
                "tests/test-data/*.prj",
                "tests/test-data/*.shp",
                "tests/test-data/*.shx",
                "tests/test-data/*.xml",
                "tests/test-data/*.tfw",
                "tests/test-data/*.tif",
                "tests/test-data/*.tif.aux.xml",
                "tests/test-data/*.tif.ovr",
                "tests/test-data/*.tif.xml",
            ]
        },
        url="https://github.com/calekochenour/awherepy",
    )
